# -*- coding: utf-8 -*-
"""
Created on Wed Aug 23 11:08:44 2023

@author: Jorrit
"""
import numpy as np
import time
import logging
import math
import matplotlib.pyplot as plt
import tkinter as tk
from scipy.signal import convolve, gaussian
from zetapy.ts_dependencies import getInterpolatedTimeSeries

# %% plottszeta2
def plottszeta2(dZETA, intPlotRandSamples=50):
    '''
    Creates figure for two-sample time-series ZETA-test

    Syntax:
    plottszeta2(dZETA, intPlotRandSamples=50)
    '''
    
    # unpack dZETA
    try:
        dblZetaP = dZETA['dblZetaP']
        dblZETA = dZETA['dblZETA']
        dblZETADeviation = dZETA['dblZETADeviation']
        dblZETATime = dZETA['dblZETATime']
        intZETAIdx = dZETA['intZETAIdx']
        dblMeanZ   = dZETA['dblMeanZ']
        dblMeanP  = dZETA['dblMeanP']
        vecMu1 = dZETA['vecMu1']
        vecMu2  = dZETA['vecMu2']
        dblZETADeviation_InvSign = dZETA['dblZETADeviation_InvSign']
        dblZETATime_InvSign = dZETA['dblZETATime_InvSign']
        intZETAIdx_InvSign = dZETA['intZETAIdx_InvSign']
        vecRefTime = dZETA['vecRefTime']
        vecRealDiff = dZETA['vecRealDiff']
        matRandDiff = dZETA['matRandDiff']
        vecRealFrac1 = dZETA['vecRealFrac1']
        vecRealFrac2 = dZETA['vecRealFrac2']
        dblUseMaxDur = dZETA['dblUseMaxDur']
        matTracePerTrial1 = dZETA['matTracePerTrial1']
        matTracePerTrial2 = dZETA['matTracePerTrial2']
    except:
        raise Exception(
            "plotzeta2 error: information is missing from dZETA dictionary")

    # %% plot
    # Plot maximally 50 traces (or however man y are requested)
    intPlotRandSamples = np.min([matRandDiff.shape[0], intPlotRandSamples])

    # Calculate optimal DPI depending on the monitor size
    screen_width = tk.Tk().winfo_screenwidth()
    dpi = screen_width / 15

    # Create figure
    f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 6), dpi=dpi)
    
    # top left: heat map 1
    x0 = vecRefTime[1]
    x1 = vecRefTime[-1]
    xw = x1-x0
    intTrialNum1 = matTracePerTrial1.shape[0]
    yh = intTrialNum1-1
    pos = ax1.imshow(matTracePerTrial1, interpolation='none', extent=[x0, x1, 1, intTrialNum1])
    ax1.set_aspect((xw/yh)/2)
    ax1.set(xlabel='Time after event (s)', ylabel='Trial number',
            title='Cond1; Color indicates data value')
    f.colorbar(pos, ax=ax1)

    # bottom left: heat map 2
    x0 = vecRefTime[1]
    x1 = vecRefTime[-1]
    xw = x1-x0
    intTrialNum2 = matTracePerTrial2.shape[0]
    yh = intTrialNum2-1
    pos = ax2.imshow(matTracePerTrial2, interpolation='none', extent=[x0, x1, 1, intTrialNum2])
    ax2.set_aspect((xw/yh)/2)
    ax2.set(xlabel='Time after event (s)', ylabel='Trial number',
            title='Cond2; Color indicates data value')
    f.colorbar(pos, ax=ax2)
    
    # top right: cumulative sums
    ax3.plot(vecRefTime, vecRealFrac1)
    ax3.plot(vecRefTime, vecRealFrac2)
    ax3.set(xlabel='Time after event (s)', ylabel='Scaled cumulative data (s)')

    # bottom right: deviation with random bootstraps
    for i in range(intPlotRandSamples-1):
        ax4.plot(vecRefTime, matRandDiff[i,:], color=[0.7, 0.7, 0.7])
    ax4.plot(vecRefTime, vecRealDiff)
    ax4.plot(dblZETATime, dblZETADeviation, 'bx')
    ax4.plot(dblZETATime_InvSign, dblZETADeviation_InvSign, 'b*')
    ax4.set(xlabel='Time after event (s)', ylabel='Difference in cumulative density')
    if dblMeanZ is not None:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f}), z(Hz)={dblMeanZ:.3f} (p={dblMeanP:.3f})')
    else:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f})')

    f.tight_layout()
    plt.show()

# %% plotzeta2
def plotzeta2(vecSpikeTimes1, arrEventTimes1, vecSpikeTimes2, arrEventTimes2, dZETA,
              intPlotRandSamples=50, intPlotSpikeNum=10000):
    '''
    Creates figure for two-sample ZETA-test

    Syntax:
    plotzeta2(vecSpikeTimes1, arrEventTimes1, vecSpikeTimes2, arrEventTimes2, dZETA,
              intPlotRandSamples=50, intPlotSpikeNum=10000)

    Parameters
    ----------
    vecSpikeTimes1 : 1D array (float)
        spike times (in seconds) for condition 1.
    arrEventTimes1 : 1D or 2D array (float)
        event on times (s) for condition 1, or [T x 2] including event off times to calculate mean-rate difference.
    vecSpikeTimes2 : 1D array (float)
        spike times (in seconds) for condition 2.
    arrEventTimes2 : 1D or 2D array (float)
        event on times (s) for condition 2, or [T x 2] including event off times to calculate mean-rate difference.
    dZETA : dict
        Output of zetatest2.
    intPlotRandSamples : int, optional
        Maximum number of random resampling to plot. The default is 50.
    intPlotSpikeNum : int, optional
        Maximum number of spikes to plot. The default is 10000.


    Code by Jorrit Montijn

    Version history:
    1.0 - 25 October 2023 Created by Jorrit Montijn
    '''
    
    # %% check input
    
    # vecSpikeTimes1 must be [S by 1] array
    assert (len(vecSpikeTimes1.shape) == 1 or vecSpikeTimes1.shape[1] == 1) and issubclass(
        vecSpikeTimes1.dtype.type, np.floating), "Input vecSpikeTimes1 is not a 1D float np.array with >2 spike times"
    vecSpikeTimes1 = np.sort(vecSpikeTimes1.flatten(), axis=0)

    # vecSpikeTimes2 must be [S by 1] array
    assert (len(vecSpikeTimes2.shape) == 1 or vecSpikeTimes2.shape[1] == 1) and issubclass(
        vecSpikeTimes2.dtype.type, np.floating), "Input vecSpikeTimes2 is not a 1D float np.array with >2 spike times"
    vecSpikeTimes2 = np.sort(vecSpikeTimes2.flatten(), axis=0)

    # ensure orientation and assert that arrEventTimes1 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes1.shape) < 3 and issubclass(
        arrEventTimes1.dtype.type, np.floating), "Input arrEventTimes1 is not a 1D or 2D float np.array"
    if len(arrEventTimes1.shape) > 1:
        if arrEventTimes1.shape[1] < 3:
            pass
        elif arrEventTimes1.shape[0] < 3:
            arrEventTimes1 = arrEventTimes1.T
        else:
            raise Exception(
                "Input error: arrEventTimes1 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes1 = np.reshape(arrEventTimes1, (-1, 1))
    # define event starts
    vecEventStarts1 = arrEventTimes1[:, 0]

    # ensure orientation and assert that arrEventTimes2 is a 1D or N-by-2 array of floats
    assert len(arrEventTimes2.shape) < 3 and issubclass(
        arrEventTimes2.dtype.type, np.floating), "Input arrEventTimes2 is not a 1D or 2D float np.array"
    if len(arrEventTimes2.shape) > 1:
        if arrEventTimes2.shape[1] < 3:
            pass
        elif arrEventTimes2.shape[0] < 3:
            arrEventTimes2 = arrEventTimes2.T
        else:
            raise Exception(
                "Input error: arrEventTimes2 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes2 = np.reshape(arrEventTimes2, (-1, 1))
    # define event starts
    vecEventStarts2 = arrEventTimes2[:, 0]

    # unpack dZETA
    try:
        dblUseMaxDur = dZETA['dblUseMaxDur']
        dblZETA = dZETA['dblZETA']
        dblZetaP = dZETA['dblZetaP']
        dblZETADeviation = dZETA['dblZETADeviation']
        dblZetaT = dZETA['dblZetaT']
        intZetaIdx = dZETA['intZetaIdx']
        
        dblD_InvSign = dZETA['dblD_InvSign']
        dblZetaT_InvSign = dZETA['dblZetaT_InvSign']
        intZetaIdx_InvSign = dZETA['intZetaIdx_InvSign']

        dblMeanZ = dZETA['dblMeanZ']
        dblMeanP = dZETA['dblMeanP']

        vecSpikeT = dZETA['vecSpikeT']
        vecRealDiff = dZETA['vecRealDiff']
        vecRealFrac1 = dZETA['vecRealFrac1']
        vecRealFrac2 = dZETA['vecRealFrac2']
        cellRandTime = dZETA['cellRandTime']
        cellRandDiff = dZETA['cellRandDiff']

    except:
        raise Exception(
            "plotzeta2 error: information is missing from dZETA dictionary")

    # %% plot
    # Plot maximally 50 traces (or however man y are requested)
    intPlotRandSamples = np.min([len(cellRandTime), intPlotRandSamples])

    # Calculate optimal DPI depending on the monitor size
    screen_width = tk.Tk().winfo_screenwidth()
    dpi = screen_width / 15

    # Create figure
    f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 6), dpi=dpi)

    # reduce spikes
    if vecSpikeTimes1.size > intPlotSpikeNum or vecSpikeTimes2.size > intPlotSpikeNum:
        dblReduceSpikesBy = min(vecSpikeTimes1.size / intPlotSpikeNum, vecSpikeTimes2.size / intPlotSpikeNum)
        intPlotSpikeNum1 = np.round(dblReduceSpikesBy * vecSpikeTimes1.size)
        intPlotSpikeNum2 = np.round(dblReduceSpikesBy * vecSpikeTimes2.size)
        vecSpikeT1_reduced = vecSpikeTimes1[np.round(np.linspace(0, vecSpikeTimes1.size-1, intPlotSpikeNum1))]
        vecSpikeT2_reduced = vecSpikeTimes1[np.round(np.linspace(0, vecSpikeTimes2.size-1, intPlotSpikeNum2))]
    else:
        vecSpikeT1_reduced = vecSpikeTimes1
        vecSpikeT2_reduced = vecSpikeTimes2

    # top left: raster 1
    for i, t in enumerate(vecEventStarts1):
        idx = np.bitwise_and(vecSpikeT1_reduced >= t, vecSpikeT1_reduced <= t + dblUseMaxDur)
        event_spks = vecSpikeT1_reduced[idx]
        ax1.vlines(event_spks - t, i + 1, i, color='k', lw=0.3)
    ax1.set(xlabel='Time after event (s)', ylabel='Trial #', title='Spike raster plot 1')

    # bottom left: raster 2
    for i, t in enumerate(vecEventStarts2):
        idx = np.bitwise_and(vecSpikeT2_reduced >= t, vecSpikeT2_reduced <= t + dblUseMaxDur)
        event_spks = vecSpikeT2_reduced[idx]
        ax3.vlines(event_spks - t, i + 1, i, color='k', lw=0.3)
    ax3.set(xlabel='Time after event (s)', ylabel='Trial #', title='Spike raster plot 2')

    
    # top right: cumulative sums
    ax2.plot(vecSpikeT, vecRealFrac1)
    ax2.plot(vecSpikeT, vecRealFrac2)
    ax2.set(xlabel='Time after event (s)', ylabel='Scaled cumulative spiking density (s)')

    # bottom right: deviation with random jitters
    for i in range(intPlotRandSamples-1):
        ax4.plot(cellRandTime[i], cellRandDiff[i], color=[0.7, 0.7, 0.7])
    ax4.plot(vecSpikeT, vecRealDiff)
    ax4.plot(dblZetaT, dblZETADeviation, 'bx')
    ax4.plot(dblZetaT_InvSign, dblD_InvSign, 'b*')
    ax4.set(xlabel='Time after event (s)', ylabel='Difference in cumulative density (s)')
    if dblMeanZ is not None:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f}), z(Hz)={dblMeanZ:.3f} (p={dblMeanP:.3f})')
    else:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f})')

    f.tight_layout()
    plt.show()

# %% plotzeta
def plotzeta(vecSpikeTimes, arrEventTimes, dZETA, dRate,
             intPlotRandSamples=50, intPlotSpikeNum=10000):
    '''
    Creates figure for ZETA-test analysis

    Syntax:
    plotzeta(vecSpikeTimes, arrEventTimes, dZETA, dRate, intPlotRandSamples=50, intPlotSpikeNum=10000)

    Parameters
    ----------
    vecSpikeTimes : 1D array (float)
        spike times (in seconds).
    arrEventTimes : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times to calculate on/off difference.
    dZETA : dict
        Output of zetatest.
    dRate : dict
        Output of zetatest.
    intPlotRandSamples : int, optional
        Maximum number of random resampling to plot. The default is 50.
    intPlotSpikeNum : int, optional
        Maximum number of spikes to plot. The default is 10000.


    Code by Jorrit Montijn

    Version history:
    1.0 - 07 September 2023 Created by Jorrit Montijn
    '''

    # %% check input
    # vecSpikeTimes must be [S by 1] array
    assert (len(vecSpikeTimes.shape) == 1 or vecSpikeTimes.shape[1] == 1) and issubclass(
        vecSpikeTimes.dtype.type, np.floating), "Input vecSpikeTimes is not a 1D float np.array with >2 spike times"
    vecSpikeTimes = np.sort(vecSpikeTimes.flatten(), axis=0)

    # ensure orientation and assert that arrEventTimes is a 1D or N-by-2 array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventStarts = arrEventTimes[:, 0]

    # unpack dZETA
    try:
        dblUseMaxDur = dZETA['dblUseMaxDur']
        dblZETA = dZETA['dblZETA']
        dblZetaP = dZETA['dblZetaP']
        dblZETADeviation = dZETA['dblZETADeviation']
        dblLatencyZETA = dZETA['dblLatencyZETA']

        dblD_InvSign = dZETA['dblD_InvSign']
        dblLatencyInvZETA = dZETA['dblLatencyInvZETA']
        intIdx_InvSign = dZETA['intIdx_InvSign']

        dblMeanZ = dZETA['dblMeanZ']
        dblMeanP = dZETA['dblMeanP']

        vecSpikeT = dZETA['vecSpikeT']
        vecRealDeviation = dZETA['vecRealDeviation']
        vecRealFrac = dZETA['vecRealFrac']
        vecRealFracLinear = dZETA['vecRealFracLinear']
        cellRandTime = dZETA['cellRandTime']
        cellRandDeviation = dZETA['cellRandDeviation']
        intZETAIdx = dZETA['intZETAIdx']

    except:
        raise Exception(
            "plotzeta error: information is missing from dZETA dictionary")

    # unpack dRate
    try:
        vecRate = dRate['vecRate']
        vecRateT = dRate['vecT']
        vecM = dRate['vecM']
        vecScale = dRate['vecScale']
        matMSD = dRate['matMSD']
        vecV = dRate['vecV']
        dblSmoothSd = dRate['dblSmoothSd']
        dblMeanRate = dRate['dblMeanRate']
    except:
        raise Exception(
            "plotzeta error: information is missing from dRate dictionary")

    # %% plot
    # Plot maximally 50 traces (or however man y are requested)
    intPlotRandSamples = np.min([len(cellRandTime), intPlotRandSamples])

    # Calculate optimal DPI depending on the monitor size
    screen_width = tk.Tk().winfo_screenwidth()
    dpi = screen_width / 15

    # Create figure
    f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 6), dpi=dpi)

    # top left: raster
    if vecSpikeTimes.size > intPlotSpikeNum:
        vecSpikeT_reduced = vecSpikeTimes[np.round(np.linspace(0, vecSpikeTimes.size-1, intPlotSpikeNum))]
    else:
        vecSpikeT_reduced = vecSpikeTimes

    for i, t in enumerate(vecEventStarts):
        idx = np.bitwise_and(vecSpikeT_reduced >= t, vecSpikeT_reduced <= t + dblUseMaxDur)
        event_spks = vecSpikeT_reduced[idx]
        ax1.vlines(event_spks - t, i + 1, i, color='k', lw=0.3)
    ax1.set(xlabel='Time after event (s)', ylabel='Trial #', title='Spike raster plot')

    # top right: psth
    peth, binned_spikes = calculatePeths(vecSpikeTimes, np.ones(vecSpikeTimes.shape), [1],
                                         vecEventStarts, pre_time=0, post_time=dblUseMaxDur,
                                         bin_size=dblUseMaxDur/25, smoothing=0)
    ax2.errorbar(peth['tscale'], peth['means'][0, :], yerr=peth['sems'])
    ax2.set(xlabel='Time after event (s)', ylabel='spks/s',
            title='Mean spiking over trials')

    # bottom left: deviation with random jitters
    for i in range(intPlotRandSamples-1):
        ax3.plot(cellRandTime[i], cellRandDeviation[i], color=[0.7, 0.7, 0.7])
    ax3.plot(vecSpikeT, vecRealDeviation)
    ax3.plot(dblLatencyZETA, dblZETADeviation, 'bx')
    ax3.plot(dblLatencyInvZETA, dblD_InvSign, 'b*')
    ax3.set(xlabel='Time after event (s)', ylabel='Spiking density anomaly (s)')
    if dblMeanZ is not None:
        ax3.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f}), z(Hz)={dblMeanZ:.3f} (p={dblMeanP:.3f})')
    else:
        ax3.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f})')

    # bottom right: ifr
    if len(vecRateT) > 1000:
        vecSubset = np.round(np.linspace(0, len(vecRateT)-1, 1000)).astype(int)
        ax4.plot(vecRateT[vecSubset], vecRate[vecSubset])
    else:
        ax4.plot(vecRateT, vecRate)

    if dblMeanRate == 1:
        strLabelY = 'Time-locked activation (a.u.)'
    else:
        strLabelY = 'Spiking rate (Hz)'
    ax4.set(xlabel='Time after event (s)', ylabel=strLabelY, title='IFR (instantaneous firing rate)')

    # plot onsets
    vecLatencies = dZETA['vecLatencies']
    vecLatencyVals = dZETA['vecLatencyVals']
    if len(vecLatencies) > 2 and vecLatencies[2] is not None and ~np.isnan(vecLatencies[2]):
        # plot peak time
        ax4.plot(vecLatencies[2], vecLatencyVals[2], 'gx')

    if len(vecLatencies) > 3 and vecLatencies[3] is not None and ~np.isnan(vecLatencies[3]):
        # plot onset time
        ax4.plot(vecLatencies[3], vecLatencyVals[3], 'rx')

    f.tight_layout()
    plt.show()

# %% plottszeta


def plottszeta(vecTime, vecData, arrEventTimes, dZETA, intPlotRandSamples=50):
    '''


    Parameters
    ----------
    vecTime [N x 1]: 1D array (float)
        timestamps in seconds corresponding to entries in vecValue.
    vecData [N x 1] : 1D array (float)
        values (e.g., dF/F0 activity).
    arrEventTimes : 1D or 2D array (float)
        event on times (s), or [T x 2] including event off times to calculate on/off difference.
    dZETA : dict
        Output of zetatstest.
    intPlotRandSamples : int, optional
        Maximum number of random resampling to plot. The default is 50.


    Code by Jorrit Montijn

    Version history:
    1.0 - 07 September 2023 Created by Jorrit Montijn
    '''

    # %% prep data and assert inputs are correct

    # vecTime and vecValue must be [N by 1] arrays
    assert len(vecTime.shape) == len(
        vecData.shape) and vecTime.shape == vecData.shape, "vecTime and vecValue have different shapes"
    assert (len(vecTime.shape) == 1 or vecTime.shape[1] == 1) and issubclass(
        vecTime.dtype.type, np.floating), "Input vecTime is not a 1D float np.array with >2 spike times"
    vecTime = vecTime.flatten()
    vecData = vecData.flatten()
    vecReorder = np.argsort(vecTime, axis=0)
    vecTime = vecTime[vecReorder]
    vecData = vecData[vecReorder]

    # ensure orientation and assert that arrEventTimes is a 1D or N-by-2 array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventTimes = arrEventTimes[:, 0]

    # check if number of events and values is sufficient
    if vecTime.size < 3 or vecEventTimes.size < 3:
        if vecTime.size < 3:
            strMsg1 = f"Number of entries in time-series ({vecTime.size}) is too few; "
        else:
            strMsg1 = ""
        if vecEventTimes.size < 3:
            strMsg2 = f"Number of events ({vecEventTimes.size}) is too few; "
        else:
            strMsg2 = ""
            logging.warning("plottszeta: " + strMsg1 + strMsg2 + "defaulting to p=1.0")

    # unpack dZETA
    try:
        # ZETA significance
        dblZetaP = dZETA['dblZetaP']
        dblZETA = dZETA['dblZETA']
        # mean-rate significance
        dblMeanZ = dZETA['dblMeanZ']
        dblMeanP = dZETA['dblMeanP']
        # data on ZETA peak
        dblZETADeviation = dZETA['dblZETADeviation']
        dblLatencyZETA = dZETA['dblLatencyZETA']
        intZETAIdx = dZETA['intZETAIdx']
        # data underlying mean-rate test
        vecMu_Dur = dZETA['vecMu_Dur']
        vecMu_Base = dZETA['vecMu_Base']
        # inverse-sign ZETA
        dblD_InvSign = dZETA['dblD_InvSign']
        dblLatencyInvZETA = dZETA['dblLatencyInvZETA']
        intIdx_InvSign = dZETA['intIdx_InvSign']

        # derived from calcZetaOne
        vecRealTime = dZETA['vecRealTime']
        vecRealDeviation = dZETA['vecRealDeviation']
        vecRealFrac = dZETA['vecRealFrac']
        vecRealFracLinear = dZETA['vecRealFracLinear']
        cellRandTime = dZETA['cellRandTime']
        cellRandDeviation = dZETA['cellRandDeviation']
        dblUseMaxDur = dZETA['dblUseMaxDur']

    except:
        raise Exception(
            "plottszeta error: information is missing from dZETA dictionary")

    # %% calculate heat map
    # sampling interval
    dblSamplingInterval = np.median(np.diff(vecTime))
    vecRefT = np.arange(dblSamplingInterval/2, dblUseMaxDur, dblSamplingInterval)
    vecRefT, matAct = getInterpolatedTimeSeries(vecTime, vecData, vecEventTimes, vecRefT)

    # %% plot
    # Plot maximally 50 traces (or however man y are requested)
    intPlotRandSamples = np.min([len(cellRandTime), intPlotRandSamples])

    # Calculate optimal DPI depending on the monitor size
    screen_width = tk.Tk().winfo_screenwidth()
    dpi = screen_width / 15

    # Create figure
    f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 6), dpi=dpi)

    # top left: heat map
    x0 = vecRefT[1]
    x1 = vecRefT[-1]
    xw = x1-x0
    intTrialNum = matAct.shape[0]
    yh = intTrialNum-1
    pos = ax1.imshow(matAct, interpolation='none', extent=[x0, x1, 1, intTrialNum])
    ax1.set_aspect((xw/yh)/2)
    ax1.set(xlabel='Time after event (s)', ylabel='Trial number',
            title='Color indicates data value')
    f.colorbar(pos, ax=ax1)

    # top right: mean +/- SEM
    vecMean = np.mean(matAct, axis=0)
    vecSem = np.std(matAct, axis=0)/np.sqrt(intTrialNum)
    ax2.errorbar(vecRefT, vecMean, yerr=vecSem)
    ax2.set(xlabel='Time after event (s)', ylabel='Data value',
            title='Mean +/- SEM over trials')

    # bottom left: cumulative plots
    vecRealTime = dZETA['vecRealTime']
    vecRealFrac = dZETA['vecRealFrac']
    vecRealFracLinear = dZETA['vecRealFracLinear']
    ax3.plot(vecRealTime, vecRealFrac)
    ax3.plot(vecRealTime, vecRealFracLinear, color=[0.7, 0.7, 0.7])
    ax3.set(xlabel='Time after event (s)', ylabel='Cumulative data', title='Time-series zeta-test')

    # bottom right: deviation with random jitters
    for i in range(intPlotRandSamples-1):
        ax4.plot(cellRandTime[i], cellRandDeviation[i], color=[0.7, 0.7, 0.7])
    ax4.plot(vecRealTime, vecRealDeviation)
    ax4.plot(dblLatencyZETA, dblZETADeviation, 'bx')
    ax4.plot(dblLatencyInvZETA, dblD_InvSign, 'b*')
    ax4.set(xlabel='Time after event (s)', ylabel='Data amplitude anomaly')
    if dblMeanZ is not None:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f}), z(Hz)={dblMeanZ:.3f} (p={dblMeanP:.3f})')
    else:
        ax4.set(title=f'ZETA={dblZETA:.3f} (p={dblZetaP:.3f})')

    f.tight_layout()
    plt.show()

def calculatePeths(
        spike_times, spike_clusters, cluster_ids, align_times, pre_time=0.2,
        post_time=0.5, bin_size=0.025, smoothing=0.025, return_fr=True):
    """
    Calcluate peri-event time histograms; return means and standard deviations
    for each time point across specified clusters
    
    Code modified from Brainbox library of the International Brain Laboratory
    https://github.com/int-brain-lab/ibllib/blob/master/brainbox/singlecell.py
    
    :param spike_times: spike times (in seconds)
    :type spike_times: array-like
    :param spike_clusters: cluster ids corresponding to each event in `spikes`
    :type spike_clusters: array-like
    :param cluster_ids: subset of cluster ids for calculating peths
    :type cluster_ids: array-like
    :param align_times: times (in seconds) to align peths to
    :type align_times: array-like
    :param pre_time: time (in seconds) to precede align times in peth
    :type pre_time: float
    :param post_time: time (in seconds) to follow align times in peth
    :type post_time: float
    :param bin_size: width of time windows (in seconds) to bin spikes
    :type bin_size: float
    :param smoothing: standard deviation (in seconds) of Gaussian kernel for
        smoothing peths; use `smoothing=0` to skip smoothing
    :type smoothing: float
    :param return_fr: `True` to return (estimated) firing rate, `False` to return spike counts
    :type return_fr: bool
    :return: peths, binned_spikes
    :rtype: peths: Bunch({'mean': peth_means, 'std': peth_stds, 'tscale': ts, 'cscale': ids})
    :rtype: binned_spikes: np.array (n_align_times, n_clusters, n_bins)
    """

    # initialize containers
    n_offset = 5 * int(np.ceil(smoothing / bin_size))  # get rid of boundary effects for smoothing
    n_bins_pre = int(np.ceil(pre_time / bin_size)) + n_offset
    n_bins_post = int(np.ceil(post_time / bin_size)) + n_offset
    n_bins = n_bins_pre + n_bins_post
    binned_spikes = np.zeros(shape=(len(align_times), len(cluster_ids), n_bins))

    # build gaussian kernel if requested
    if smoothing > 0:
        w = n_bins - 1 if n_bins % 2 == 0 else n_bins
        window = gaussian(w, std=smoothing / bin_size)
        # half (causal) gaussian filter
        # window[int(np.ceil(w/2)):] = 0
        window /= np.sum(window)
        binned_spikes_conv = np.copy(binned_spikes)

    ids = np.unique(cluster_ids)

    # filter spikes outside of the loop
    idxs = np.bitwise_and(spike_times >= np.min(align_times) - (n_bins_pre + 1) * bin_size,
                          spike_times <= np.max(align_times) + (n_bins_post + 1) * bin_size)
    idxs = np.bitwise_and(idxs, np.isin(spike_clusters, cluster_ids))
    spike_times = spike_times[idxs]
    spike_clusters = spike_clusters[idxs]

    # compute floating tscale
    tscale = np.arange(-n_bins_pre, n_bins_post + 1) * bin_size
    # bin spikes
    for i, t_0 in enumerate(align_times):
        # define bin edges
        ts = tscale + t_0
        # filter spikes
        idxs = np.bitwise_and(spike_times >= ts[0], spike_times <= ts[-1])
        i_spikes = spike_times[idxs]
        i_clusters = spike_clusters[idxs]

        # bin spikes similar to bincount2D: x = spike times, y = spike clusters
        xscale = ts
        xind = (np.floor((i_spikes - np.min(ts)) / bin_size)).astype(np.int64)
        yscale, yind = np.unique(i_clusters, return_inverse=True)
        nx, ny = [xscale.size, yscale.size]
        ind2d = np.ravel_multi_index(np.c_[yind, xind].transpose(), dims=(ny, nx))
        r = np.bincount(ind2d, minlength=nx * ny, weights=None).reshape(ny, nx)

        # store (ts represent bin edges, so there are one fewer bins)
        bs_idxs = np.isin(ids, yscale)
        binned_spikes[i, bs_idxs, :] = r[:, :-1]

        # smooth
        if smoothing > 0:
            idxs = np.where(bs_idxs)[0]
            for j in range(r.shape[0]):
                binned_spikes_conv[i, idxs[j], :] = convolve(
                    r[j, :], window, mode='same', method='auto')[:-1]

    # average
    if smoothing > 0:
        binned_spikes_ = np.copy(binned_spikes_conv)
    else:
        binned_spikes_ = np.copy(binned_spikes)
    if return_fr:
        binned_spikes_ /= bin_size

    peth_means = np.mean(binned_spikes_, axis=0)
    peth_stds = np.std(binned_spikes_, axis=0)
    peth_sems = np.std(binned_spikes_, axis=0) / np.sqrt(align_times.shape[0])

    if smoothing > 0:
        peth_means = peth_means[:, n_offset:-n_offset]
        peth_stds = peth_stds[:, n_offset:-n_offset]
        binned_spikes = binned_spikes[:, :, n_offset:-n_offset]
        tscale = tscale[n_offset:-n_offset]

    # package output
    tscale = (tscale[:-1] + tscale[1:]) / 2
    peths = dict({'means': peth_means, 'stds': peth_stds, 'sems': peth_sems,
                  'tscale': tscale, 'cscale': ids})
    return peths, binned_spikes


# -*- coding: utf-8 -*-
import numpy as np
import logging
from math import factorial
from zetapy.dependencies import (findfirst, getZetaP)

# %%


def calcTsZetaTwo(vecTimestamps1, vecData1, arrEventTimes1, vecTimestamps2, vecData2, arrEventTimes2, dblSuperResFactor, dblUseMaxDur, intResampNum, boolDirectQuantile):
    """
   Calculates neuronal responsiveness index zeta
    dZETA = calcTsZetaTwo(vecTimestamps1, vecData1, arrEventTimes1, vecTimestamps2, vecData2, arrEventTimes2, 
                          dblSuperResFactor, dblUseMaxDur, intResampNum, boolDirectQuantile)
    dZETA has entries:
        vecRefTime, vecRealDiff, vecRealFrac1, vecRealFrac2, matRandDiff, dblZetaP, dblZETA, intZETAIdx
    """

    # %% pre-allocate output
    vecRefTime = None
    vecRealDiff = None
    vecRealFrac1 = None
    vecRealFrac2 = None
    matRandDiff = None
    dblZetaP = 1.0
    dblZETA = 0.0
    intZETAIdx = None
    matTracePerTrial1 = None
    matTracePerTrial2 = None
   
    dZETA = dict()
    dZETA['vecRefTime'] = vecRefTime
    dZETA['vecRealDiff'] = vecRealDiff
    dZETA['vecRealFrac1'] = vecRealFrac1
    dZETA['vecRealFrac2'] = vecRealFrac2
    dZETA['matRandDiff'] = matRandDiff
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx
    dZETA['matTracePerTrial1'] = matTracePerTrial1
    dZETA['matTracePerTrial2'] = matTracePerTrial2

    # %% reduce data
    # assert that arrEventTimes is a 1D array of floats
    assert len(arrEventTimes1.shape) < 3 and len(arrEventTimes2.shape) < 3 \
        and issubclass(arrEventTimes1.dtype.type, np.floating) and issubclass(arrEventTimes2.dtype.type, np.floating), \
        "Input arrEventTimes1 or arrEventTimes2 is not a 1D or 2D float np.array"

    # ensure orientation arrEventTimes1
    if len(arrEventTimes1.shape) > 1:
        if arrEventTimes1.shape[1] < 3:
            pass
        elif arrEventTimes1.shape[0] < 3:
            arrEventTimes1 = arrEventTimes1.T
        else:
            raise Exception(
                "Input error: arrEventTimes1 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes1 = np.reshape(arrEventTimes1, (-1, 1))

    # ensure orientation arrEventTimes2
    if len(arrEventTimes2.shape) > 1:
        if arrEventTimes2.shape[1] < 3:
            pass
        elif arrEventTimes2.shape[0] < 3:
            arrEventTimes2 = arrEventTimes2.T
        else:
            raise Exception(
                "Input error: arrEventTimes2 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes2 = np.reshape(arrEventTimes2, (-1, 1))

    # reduce data 1
    vecEventStarts1 = arrEventTimes1[:, 0]
    dblPreUse = -dblUseMaxDur
    dblPostUse = dblUseMaxDur*2
    dblStartT1 = np.min(vecEventStarts1) + dblPreUse*2
    dblStopT1 = np.max(vecEventStarts1) + dblPostUse*2

    indKeepEntries1 = np.logical_and(vecTimestamps1 >= dblStartT1, vecTimestamps1 <= dblStopT1)
    vecTimestamps1 = vecTimestamps1[indKeepEntries1]
    vecData1 = vecData1[indKeepEntries1]

    if vecTimestamps1.size < 3:
        logging.warning(
            "calcTsZetaTwo:vecTimestamps1: too few entries around events to calculate zeta")
        return dZETA

    # reduce data 2
    vecEventStarts2 = arrEventTimes2[:, 0]
    dblPreUse = -dblUseMaxDur
    dblPostUse = dblUseMaxDur*2
    dblStartT2 = np.min(vecEventStarts2) + dblPreUse*2
    dblStopT2 = np.max(vecEventStarts2) + dblPostUse*2

    indKeepEntries2 = np.logical_and(vecTimestamps2 >= dblStartT2, vecTimestamps2 <= dblStopT2)
    vecTimestamps2 = vecTimestamps2[indKeepEntries2]
    vecData2 = vecData2[indKeepEntries2]

    if vecTimestamps2.size < 3:
        logging.warning(
            "calcTsZetaTwo:vecTimestamps2: too few entries around events to calculate zeta")
        return dZETA

    # %% rescale
    dblMin = min(np.min(vecData1), np.min(vecData2))
    dblMax = max(np.max(vecData1), np.max(vecData2))
    dblRange = (dblMax-dblMin)
    if dblRange == 0:
        dblRange = 1
        logging.warning(
            "calcTsZetaTwo:ZeroVar: Input data has zero variance")

    vecTraceAct1 = np.divide(vecData1-dblMin, dblRange)
    vecTraceAct2 = np.divide(vecData2-dblMin, dblRange)

    # %% build reference time and matrices
    # time
    vecRefT1 = getTsRefT(vecTimestamps1, vecEventStarts1, dblUseMaxDur)
    vecRefT2 = getTsRefT(vecTimestamps2, vecEventStarts2, dblUseMaxDur)
    # set tol
    dblSampInterval = (np.median(np.diff(vecRefT1)) + np.median(np.diff(vecRefT2)))/2.0
    dblTol = dblSampInterval/dblSuperResFactor
    vecRefTime = uniquetol(np.concatenate((vecRefT1, vecRefT2), axis=0), dblTol)
    intT = len(vecRefTime)

    # matrices
    vecTime1, matTracePerTrial1 = getInterpolatedTimeSeries(vecTimestamps1, vecTraceAct1, vecEventStarts1, vecRefTime)
    vecTime2, matTracePerTrial2 = getInterpolatedTimeSeries(vecTimestamps2, vecTraceAct2, vecEventStarts2, vecRefTime)

    # %% get trial responses
    vecRealDiff, vecRealFrac1, vecRealFrac2 = getTimeseriesOffsetTwo(matTracePerTrial1, matTracePerTrial2)
    intZETAIdx = np.argmax(np.abs(vecRealDiff))
    dblMaxD = np.abs(vecRealDiff[intZETAIdx])

    # repeat procedure, but swap trials randomly in each resampling
    matRandDiff = np.empty((intResampNum, intT))
    matRandDiff.fill(np.nan)
    vecMaxRandD = np.empty((intResampNum, 1))
    vecMaxRandD.fill(np.nan)

    matAggregateTrials = np.concatenate((matTracePerTrial1, matTracePerTrial2), axis=0)
    intTrials1 = matTracePerTrial1.shape[0]
    intTrials2 = matTracePerTrial2.shape[0]
    intTotTrials = intTrials1+intTrials2

    # %% run resamplings
    for intResampling in range(intResampNum):
        # %% get random subsample
        # if cond1 has 10 trials, and cond2 has 100, then:
        # for shuffle of cond1: take 10 trials from set of 110
        # for shuffle of cond2: take 100 trials from set of 110
        vecUseRand1 = np.random.randint(intTotTrials, size=intTrials1)
        vecUseRand2 = np.random.randint(intTotTrials, size=intTrials2)

        matTrace1_Rand = matAggregateTrials[vecUseRand1, :]
        matTrace2_Rand = matAggregateTrials[vecUseRand2, :]

        # get difference
        vecRandDiff, vecRandFrac1, vecRandFrac2 = getTimeseriesOffsetTwo(matTrace1_Rand, matTrace2_Rand)

        # assign data
        matRandDiff[intResampling, :] = vecRandDiff
        dblAddVal = np.max(np.abs(vecRandDiff))
        if dblAddVal is None or dblAddVal == 0:
            dblAddVal = dblMaxD
        vecMaxRandD[intResampling] = dblAddVal

    # %% calculate significance
    dblZetaP, dblZETA = getZetaP(dblMaxD, vecMaxRandD, boolDirectQuantile)
    
    # %% assign output
    dZETA['vecRefTime'] = vecRefTime
    dZETA['vecRealDiff'] = vecRealDiff
    dZETA['vecRealFrac1'] = vecRealFrac1
    dZETA['vecRealFrac2'] = vecRealFrac2
    dZETA['matRandDiff'] = matRandDiff
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx
    dZETA['matTracePerTrial1'] = matTracePerTrial1
    dZETA['matTracePerTrial2'] = matTracePerTrial2
    
    return dZETA

# %% getTimeseriesOffsetTwo


def getTimeseriesOffsetTwo(matTracePerTrial1, matTracePerTrial2):
    '''
    vecThisDiff,vecThisFrac1,vecThisFrac2 = getTimeseriesOffsetTwo(matTracePerTrial1,matTracePerTrial2)
    '''

    # cond1 goes to sum(v_mu1); cond2 goes to sum(v_mu2)
    vecMeanTrace1 = np.nanmean(matTracePerTrial1.T, axis=1)
    vecMeanTrace2 = np.nanmean(matTracePerTrial2.T, axis=1)

    # get real cumsums
    vecThisFrac1 = np.cumsum(vecMeanTrace1)
    vecThisFrac2 = np.cumsum(vecMeanTrace2)

    # take difference
    vecDeviation = vecThisFrac1 - vecThisFrac2

    # mean-subtract
    vecThisDiff = vecDeviation - np.mean(vecDeviation)

    # return
    return vecThisDiff, vecThisFrac1, vecThisFrac2

# %%


def calcTsZetaOne(vecTimestamps, vecData, arrEventTimes, dblUseMaxDur, intResampNum, boolDirectQuantile, dblJitterSize, boolStitch):
    """
   Calculates neuronal responsiveness index zeta
    dZETA = calcTsZetaOne(vecTimestamps, vecData, arrEventTimes, dblUseMaxDur, intResampNum, boolDirectQuantile, dblJitterSize, boolStitch)
    dZETA has entries:
        vecRealTime, vecRealDeviation, vecRealFrac, vecRealFracLinear, cellRandTime, cellRandDeviation, dblZetaP, dblZETA, intZETAIdx
    """

    # %% pre-allocate output
    vecRealTime = None
    vecRealDeviation = None
    vecRealFrac = None
    vecRealFracLinear = None
    cellRandTime = None
    cellRandDeviation = None
    dblZetaP = 1.0
    dblZETA = 0.0
    intZETAIdx = None

    dZETA = dict()
    dZETA['vecRealTime'] = vecRealTime
    dZETA['vecRealDeviation'] = vecRealDeviation
    dZETA['vecRealFrac'] = vecRealFrac
    dZETA['vecRealFracLinear'] = vecRealFracLinear
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDeviation'] = cellRandDeviation
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx

    # %% reduce data
    # ensure orientation and assert that arrEventTimes is a 1D array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventT = arrEventTimes[:, 0]

    dblMinPreEventT = np.min(vecEventT)-dblUseMaxDur*5*dblJitterSize
    dblStartT = max([vecTimestamps[0], dblMinPreEventT])
    dblStopT = max(vecEventT)+dblUseMaxDur*5*dblJitterSize
    indKeepEntries = np.logical_and(vecTimestamps >= dblStartT, vecTimestamps <= dblStopT)
    vecTimestamps = vecTimestamps[indKeepEntries]
    vecData = vecData[indKeepEntries]

    if vecTimestamps.size < 3:
        logging.warning(
            "calcTsZetaOne:vecTimestamps: too few entries around events to calculate zeta")
        return dZETA

    # %% build pseudo data, stitching stimulus periods
    vecPseudoT, vecPseudoV, vecPseudoEventT = getPseudoTimeSeries(vecTimestamps, vecData, vecEventT, dblUseMaxDur)
    vecPseudoV = vecPseudoV - np.min(vecPseudoV)
    if vecTimestamps.size < 3:
        logging.warning(
            "calcTsZetaOne:vecPseudoT: too few entries around events to calculate zeta")
        return dZETA

    if boolStitch:
        vecPseudoT, vecPseudoV, vecPseudoEventT = getPseudoTimeSeries(vecTimestamps, vecData, vecEventT, dblUseMaxDur)
    else:
        vecPseudoT = vecTimestamps
        vecPseudoV = vecData
        vecPseudoEventT = vecEventT

    # %% run normal
    # get data
    vecRealDeviation, vecRealFrac, vecRealFracLinear, vecRealTime = getTimeseriesOffsetOne(
        vecPseudoT, vecPseudoV, vecPseudoEventT, dblUseMaxDur)

    if vecRealDeviation.size < 3:
        logging.warning(
            "calcZetaOne:vecRealDeviation: too few spikes around events to calculate zeta")
        return dZETA

    vecRealDeviation = vecRealDeviation - np.mean(vecRealDeviation)
    intZETAIdx = np.argmax(np.abs(vecRealDeviation))
    dblMaxD = np.abs(vecRealDeviation[intZETAIdx])

    # %% create random jitters
    # run pre-set number of iterations
    cellRandTime = []
    cellRandDeviation = []
    vecMaxRandD = np.empty((intResampNum, 1))
    vecMaxRandD.fill(np.nan)

    vecStartOnly = np.reshape(vecPseudoEventT, (-1, 1))
    intTrials = vecStartOnly.size
    matJitterPerTrial = np.empty((intTrials, intResampNum))
    matJitterPerTrial.fill(np.nan)
    # uniform jitters between dblJitterSize*[-tau, +tau]
    for intResampling in range(intResampNum):
        matJitterPerTrial[:, intResampling] = dblJitterSize*dblUseMaxDur * \
            ((np.random.rand(vecStartOnly.shape[0]) - 0.5) * 2)

    # %% this part is only to check if matlab and python give the same exact results
    # unfortunately matlab's randperm() and numpy's np.random.permutation give different outputs even with
    # identical seeds and identical random number generators, so I've had to load in a table of random values here...
    boolTest = False
    if boolTest:
        from scipy.io import loadmat
        print('Loading deterministic jitter data for comparison with matlab')
        logging.warning(
            "calcZetaOne:debugMode: set boolTest to False to suppress this warning")
        dLoad = loadmat('matJitterPerTrialTsZeta.mat')
        matJitterPerTrial = dLoad['matJitterPerTrial']

        # reset rng
        np.random.seed(1)

    # %% run resamplings
    for intResampling in range(intResampNum):
        # get random subsample
        vecStimUseOnTime = vecStartOnly[:, 0] + matJitterPerTrial[:, intResampling].T

        # get temp offset
        vecRandDeviation, vecThisFrac, vecThisFracLinear, vecRandT = getTimeseriesOffsetOne(
            vecPseudoT, vecPseudoV, vecStimUseOnTime, dblUseMaxDur)

        # assign data
        cellRandTime.append(vecRandT)
        cellRandDeviation.append(vecRandDeviation - np.mean(vecRandDeviation))
        vecMaxRandD[intResampling] = np.max(np.abs(cellRandDeviation[intResampling]))

    # %% calculate significance
    dblZetaP, dblZETA = getZetaP(dblMaxD, vecMaxRandD, boolDirectQuantile)

    # %% assign output
    dZETA = dict()
    dZETA['vecRealTime'] = vecRealTime
    dZETA['vecRealDeviation'] = vecRealDeviation
    dZETA['vecRealFrac'] = vecRealFrac
    dZETA['vecRealFracLinear'] = vecRealFracLinear
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDeviation'] = cellRandDeviation
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx

    return dZETA

# %% getpseudotimeseries


def getPseudoTimeSeries(vecTimestamps, vecData, vecEventTimes, dblWindowDur):
    '''
    vecPseudoTime, vecPseudoData, vecPseudoEventT = getPseudoTimeSeries(vecTime, vecData, vecEventTimes, dblWindowDur)

    Parameters
    ----------
    vecTimestamps : TYPE
        DESCRIPTION.
    vecData : TYPE
        DESCRIPTION.
    vecEventTimes : TYPE
        DESCRIPTION.
    dblWindowDur : TYPE
        DESCRIPTION.

    Returns
    -------
    vecPseudoTime, vecPseudoData, vecPseudoEventT.

    '''
    # %% prep
    # ensure sorting and alignment
    vecTimestamps = np.squeeze(np.vstack(vecTimestamps))
    vecData = np.squeeze(np.vstack(vecData))
    vecReorder = np.argsort(vecTimestamps, axis=0)
    vecTimestamps = vecTimestamps[vecReorder]
    vecData = vecData[vecReorder]
    vecEventTimes = np.squeeze(np.sort(np.vstack(vecEventTimes), axis=0))

    # %% pre-allocate
    intSamples = vecTimestamps.size
    intTrials = vecEventTimes.size
    dblMedianDur = np.median(np.diff(vecTimestamps, axis=0))
    cellPseudoTime = []
    cellPseudoData = []
    vecPseudoEventT = np.empty((intTrials, 1))
    vecPseudoEventT.fill(np.nan)
    dblPseudoEventT = 0.0
    dblStartNextAtT = 0
    intLastUsedSample = 0
    intFirstSample = None

    # %% run
    for intTrial, dblEventT in enumerate(vecEventTimes):
        # %%
        # intTrial = intTrial + 1
        # dblEventT = vecEventTimes[intTrial]
        # get eligible samples
        intStartSample = findfirst(vecTimestamps >= dblEventT)
        intEndSample = findfirst(vecTimestamps > (dblEventT+dblWindowDur))

        if intStartSample is not None and intEndSample is not None and intStartSample > intEndSample:
            intEndSample = None
            intStartSample = None

        if intEndSample is None:
            intEndSample = len(vecTimestamps)

        if intStartSample is None or intEndSample is None:
            vecUseSamples = None
        else:
            vecEligibleSamples = np.arange(intStartSample, intEndSample)
            indUseSamples = np.logical_and(vecEligibleSamples >= 0, vecEligibleSamples < intSamples)
            vecUseSamples = vecEligibleSamples[indUseSamples]

        # check if beginning or end
        if vecUseSamples.size > 0:
            if intTrial == 0:
                vecUseSamples = np.arange(0, vecUseSamples[-1]+1)
            elif intTrial == (intTrials-1):
                vecUseSamples = np.arange(vecUseSamples[0], intSamples)

        # add entries
        vecUseT = vecTimestamps[vecUseSamples]
        indOverlap = vecUseSamples <= intLastUsedSample

        # get event t
        if intTrial == 0:
            dblPseudoEventT = 0.0
        else:
            if intTrial > 0 and dblWindowDur > (dblEventT - vecEventTimes[intTrial-1]):
                # remove spikes from overlapping epochs
                vecUseSamples = vecUseSamples[~indOverlap]
                vecUseT = vecTimestamps[vecUseSamples]
                dblPseudoEventT = dblPseudoEventT + dblEventT - vecEventTimes[intTrial-1]
            else:
                dblPseudoEventT = dblPseudoEventT + dblWindowDur

        # make local pseudo event time
        if vecUseSamples.size == 0:
            vecLocalPseudoT = None
            vecLocalPseudoV = None
            dblPseudoEventT = dblEventT - vecTimestamps[intLastUsedSample] + dblStartNextAtT
        else:
            intLastUsedSample = vecUseSamples[-1]
            vecLocalPseudoV = vecData[vecUseSamples]
            vecLocalPseudoT = vecUseT - vecUseT[0] + dblStartNextAtT
            dblPseudoEventT = dblEventT - vecUseT[0] + dblStartNextAtT

            if len(vecTimestamps) > (intLastUsedSample+1):
                dblStepEnd = vecTimestamps[intLastUsedSample+1] - vecTimestamps[intLastUsedSample]
            else:
                dblStepEnd = dblMedianDur

            dblStartNextAtT = vecLocalPseudoT[-1] + dblStepEnd

        if intFirstSample is None and vecUseSamples.size > 0:
            intFirstSample = vecUseSamples[0]
            dblPseudoT0 = dblPseudoEventT

        if vecLocalPseudoT is not None:
            # assign data for this trial
            cellPseudoTime.append(vecLocalPseudoT)
            cellPseudoData.append(vecLocalPseudoV)
            vecPseudoEventT[intTrial] = dblPseudoEventT

    # %% add beginning
    dblT1 = vecTimestamps[intFirstSample]
    intT0 = findfirst(vecTimestamps > (dblT1 - dblWindowDur))
    if intT0 is not None and intFirstSample is not None and intFirstSample > 0:
        dblStepBegin = vecTimestamps[intFirstSample] - vecTimestamps[intFirstSample-1]
        vecSampAddBeginning = np.arange(intT0-1, intFirstSample)
        vecAddBeginningT = vecTimestamps[vecSampAddBeginning] - vecTimestamps[vecSampAddBeginning[0]] \
            + dblPseudoT0 - dblStepBegin - \
            np.ptp(vecTimestamps[vecSampAddBeginning]
                   )  # make local to first entry in array, then preceding pseudo event t0
        cellPseudoTime.append(vecAddBeginningT)
        cellPseudoData.append(vecData[vecSampAddBeginning])

    # %% add end
    intFindTail = findfirst(vecTimestamps > (vecEventTimes[-1]+dblWindowDur))
    if intFindTail is None:
        raise Exception(
            "zetatstest error - dblMaxDur is too large: the tail of the final event would extend beyond the end of the time-series data. Please include more data, shorten dblMaxDur or remove the last event.")
    else:
        dblTn = vecTimestamps[intLastUsedSample]
        intTn = findfirst(vecTimestamps > dblTn)
        if intTn is not None and (intTn-1) > intLastUsedSample:
            vecSampAddEnd = np.arange(intLastUsedSample, intTn)+1
            cellPseudoTime.append(vecTimestamps[vecSampAddEnd] - vecTimestamps[vecSampAddEnd[0]] + dblStartNextAtT)
            cellPseudoData.append(vecData[vecSampAddEnd])

    # %% recombine into vector
    vecPseudoTime = np.vstack(np.concatenate(cellPseudoTime))
    vecPseudoData = np.vstack(np.concatenate(cellPseudoData))
    return vecPseudoTime, vecPseudoData, vecPseudoEventT


# %% getTimeseriesOffsetOne
def getTimeseriesOffsetOne(vecTimestamps, vecData, vecEventStartT, dblUseMaxDur):
    '''
    vecDeviation, vecThisFrac, vecThisFracLinear, vecTime = getTimeseriesOffsetOne(vecT, vecV, vecEventT, dblUseMaxDur)

    Parameters
    ----------
    vecTimestamps : TYPE
        DESCRIPTION.
    vecData : TYPE
        DESCRIPTION.
    vecEventStartT : TYPE
        DESCRIPTION.
    dblUseMaxDur : TYPE
        DESCRIPTION.

    Returns
    -------
    vecDeviation, vecThisFrac, vecThisFracLinear, vecTime.

    '''

    # %% prepare
    vecTime = getTsRefT(vecTimestamps, vecEventStartT, dblUseMaxDur)

    # build interpolated data
    vecTime, matTracePerTrial = getInterpolatedTimeSeries(vecTimestamps, vecData, vecEventStartT, vecTime)
    indKeepPoints = np.logical_and(vecTime >= 0, vecTime <= dblUseMaxDur)
    vecTime = vecTime[indKeepPoints]
    matTracePerTrial = matTracePerTrial[:, indKeepPoints]
    vecMeanTrace = np.nanmean(matTracePerTrial, axis=0)
    vecThisFrac = np.cumsum(vecMeanTrace) / np.sum(vecMeanTrace)

    # get linear fractions
    vecThisFracLinear = np.linspace(np.mean(vecMeanTrace), np.sum(
        vecMeanTrace), len(vecMeanTrace)) / np.sum(vecMeanTrace)

    # assign data
    vecDeviation = vecThisFrac - vecThisFracLinear
    vecDeviation = vecDeviation - np.mean(vecDeviation)

    # %% return
    return vecDeviation, vecThisFrac, vecThisFracLinear, vecTime

# %% getTsRefT


def getTsRefT(vecTimestamps, vecEventStartT, dblUseMaxDur):
    # pre-allocate
    vecEventStartT = np.sort(vecEventStartT)
    intTimeNum = len(vecTimestamps)-1

    # intTrial = -1
    # build common timeframe
    cellRefT = []
    for intTrial, dblStartT in enumerate(vecEventStartT):
        # %%
        # intTrial = intTrial + 1
        # dblStartT = vecEventStartT[intTrial]
        # get original times
        intBegin = findfirst(vecTimestamps > dblStartT)
        if intBegin is None:
            intStartT = 0
        else:
            intStartT = np.max([0, intBegin - 1])

        dblStopT = dblStartT+dblUseMaxDur
        intEnd = findfirst(vecTimestamps > dblStopT)
        if intEnd is None:
            intStopT = intTimeNum
        else:
            intStopT = np.min([intTimeNum, intEnd])

        vecSelectSamples = np.arange(intStartT, intStopT+1)

        # save data
        cellRefT.append(vecTimestamps[vecSelectSamples]-dblStartT)

    # %% set tol
    dblSampInterval = np.median(np.diff(vecTimestamps, axis=0))
    dblTol = dblSampInterval/100
    vecVals = np.sort(np.vstack(np.concatenate(cellRefT)))
    vecTime = np.hstack(uniquetol(vecVals, dblTol))

    # return
    return vecTime

# %% getInterpolatedTimeSeries


def getInterpolatedTimeSeries(vecTimestamps, vecData, vecEventStartT, vecRefTime):
    '''
   getInterpolatedTimeSeries Builds common timeframe
     syntax: vecRefTime,matTracePerTrial = getInterpolatedTimeSeries(vecTimestamps,vecData,vecEventStartT,dblUseMaxDur,vecRefTime)
       input:
       - vecTimestamps; time stamps (s)
       - vecData; time-series data
       - vecEventStartT: trial start times (s)
       - dblUseMaxDur: window (s)
       - vecTime: reference time vector (s)

     Version history:
     1.0 - June 26 2019
         Created by Jorrit Montijn
         '''

    # assign data
    vecRefTime = np.hstack((vecRefTime))
    vecTimestamps = np.hstack((vecTimestamps))
    vecData = np.hstack((vecData))
    matTracePerTrial = np.zeros((len(vecEventStartT), len(vecRefTime)))
    for intTrial, dblStartT in enumerate(vecEventStartT):
        # original times
        intBegin = findfirst(vecTimestamps > (dblStartT + vecRefTime[0]))
        if intBegin is None:
            raise Exception(
                "getInterpolatedTimeSeries error - no time stamps exist after trial start")

        intStartT = np.max([0, intBegin - 1])
        intEnd = findfirst(vecTimestamps > (dblStartT + vecRefTime[-1]))
        if intEnd is None:
            intStopT = len(vecTimestamps)
        else:
            intStopT = np.min([len(vecTimestamps), intEnd + 1])

        vecSelectSamples = np.arange(intStartT, intStopT)

        # get data
        vecUseTimes = vecTimestamps[vecSelectSamples]
        vecUseData = vecData[vecSelectSamples]

        # interpolate to
        vecUseInterpT = vecRefTime+dblStartT

        # get interpolated data
        matTracePerTrial[intTrial, :] = np.interp(vecUseInterpT, vecUseTimes, vecUseData)

    # return
    return vecRefTime, matTracePerTrial

# %% uniquetol


def uniquetol(array_in, dblTol):
    '''
    array_unique = uniquetol(array_in,dblTol)

    Parameters
    ----------
    array_in : np.array (float)
    dblTol : tolerance (float)

    Returns
    -------
    array_unique : np.array (float)
        array with unique values within tolerance.

    '''
    return (np.unique(np.floor(array_in/dblTol).astype(int)))*dblTol



from zetapy.main import zetatest, zetatest2, zetatstest, zetatstest2, ifr
from zetapy.plot_dependencies import plotzeta, plotzeta2, plottszeta, plottszeta2
from zetapy import legacy
from zetapy.legacy.main import getZeta, getIFR

# from zetapy.msd import getMultiScaleDeriv
# from zetapy.dependencies import getPeak, getGumbel, getOnset, getTempOffset, flatten, calculatePeths



import logging
from math import log
from scipy.stats import norm
from scipy.signal import convolve2d
import numpy as np


def getMultiScaleDeriv(vecT, vecV, intSmoothSd=0, dblMinScale=None, dblBase=1.5, intPlot=0,
                       dblMeanRate=1, dblUseMaxDur=None, axs=None):
    """Returns multi-scale derivative. Syntax:
       [vecRate,sMSD] = getMultiScaleDeriv(vecT,vecV,intSmoothSd,dblMinScale,dblBase,intPlot,dblMeanRate,dblUseMaxDur)
    Required input:
        - vecT [N x 1]: timestamps (e.g., spike times)
        - vecV [N x 1]: values (e.g., z-scores)

    Optional inputs:
        - intSmoothSd: Gaussian SD of smoothing kernel (in # of samples) [default: 0]
        - dblMinScale: minimum derivative scale in seconds [default: 1/1000]
        - dblBase: base for exponential scale step size [default: 1.5]
        - intPlot: integer, plotting switch 
        - dblMeanRate: mean spiking rate to normalize vecRate (optional)
        - dblUseMaxDur: trial duration to normalize vecRate (optional)
        - axs: two element vector with the handles to the two subplots that are to be plotted

    Outputs:
        - vecRate; Instantaneous spiking rate
        - sMSD; structure with fields:
            - vecRate; instantaneous spiking rates (like a PSTH)
            - vecT; time-points corresponding to vecRate (same as input vecT)
            - vecM; Mean of multi-scale derivatives
            - vecScale; timescales used to calculate derivatives
            - matMSD; multi-scale derivatives matrix
            - vecV; values on which vecRate is calculated (same as input vecV)

    Version history:
    %1.1 - June 18, 2020 created by Jorrit Montijn, translated to python by Alexander Heimel
    """

    ## set default values
    if dblMinScale==None:
        dblMinScale = round(log(1/1000) / log(dblBase))

    if dblUseMaxDur == None:
        dblUseMaxDur = np.max(vecT) - np.min(vecT)

    ## reorder just in case
    ### [vecT,vecReorder] = sort(vecT(:),'ascend');
    ### vecV = vecV(vecReorder);
    ### vecV = vecV(:);
    vecReorder = np.argsort(vecT)
    vecT = np.array(vecT)[vecReorder]
    vecV = np.array(vecV)[vecReorder]

    ## prepare data
    dblMaxScale = log(np.max(vecT)/10) / log(dblBase)
    intN = len(vecT)

    ## get multi-scale derivative
    ### vecExp = dblMinScale:dblMaxScale;
    vecExp = np.arange(dblMinScale,dblMaxScale+1)
    vecScale = [dblBase ** x for x in vecExp]
    intScaleNum = len(vecScale)
    matMSD = np.zeros( (intN,intScaleNum) )

    #logging.warning('Loop parallelization is not yet translated to python.')
    # try %try parallel
    #     parfor intScaleIdx=1:intScaleNum
    #         dblScale = vecScale(intScaleIdx);

    #         %run through all points
    #         for intS=1:intN
    #             %select points within window
    #             matMSD(intS,intScaleIdx) = getD(dblScale,intS,intN,vecT,vecV);
    #         end
    #     end
    # catch %otherwise try normal loop
    for intScaleIdx in range(intScaleNum):
        dblScale = vecScale[intScaleIdx]
        # run through all points
        for intS in range(intN):
            # select points within window
            matMSD[intS,intScaleIdx] = getD(dblScale, intS, intN, vecT, vecV)

    ## smoothing
    if intSmoothSd > 0:
        ### vecFilt = normpdf(-2*(intSmoothSd):2*intSmoothSd,0,intSmoothSd)';
        vecFilt = norm.pdf( range(-2*(intSmoothSd),2*intSmoothSd+1), 0, intSmoothSd)
        vecFilt = vecFilt / sum(vecFilt)
        # pad array
        ### matMSD = padarray(matMSD,floor(size(vecFilt)/2),'replicate');
        x = int(np.floor(len(vecFilt)/2))
        matMSD = np.pad(matMSD,( (x,x),(0,0)),'edge')

        # filter
        ### matMSD = conv2(matMSD,vecFilt,'valid');
        matMSD = convolve2d(matMSD, np.transpose([vecFilt]), 'valid')

    # mean
    ### vecM = mean(matMSD,2);
    vecM = np.mean(matMSD,1)

    # weighted average of vecM by inter-spike intervals
    ### dblMeanM = (1/dblUseMaxDur) * sum(((vecM(1:(end-1)) + vecM(2:end))/2).*diff(vecT));
    dblMeanM = (1/dblUseMaxDur) * sum(((vecM[:-1] + vecM[1:])/2) * np.diff(vecT))

    # rescale to real firing rates
    ### vecRate = dblMeanRate * ((vecM + 1/dblUseMaxDur)/(dblMeanM + 1/dblUseMaxDur));
    vecRate = dblMeanRate * ((vecM + 1/dblUseMaxDur) / (dblMeanM + 1/dblUseMaxDur))

    ## plot
    if intPlot == 1:
        if intSmoothSd > 0:
            strTitle = 'Smoothed MSDs'
        else:
            strTitle = 'MSDs'
        axs[0].imshow(matMSD.T, aspect='auto')
        axs[0].set(yticks=[], ylabel='Scale (s)',
                   xlabel='Timestamp index (#)', title=strTitle)    
        
        if len(vecT) > 1000:
            vecSubset = np.round(np.linspace(0, len(vecT)-1, 1000)).astype(int)
            axs[1].plot(vecT[vecSubset], vecRate[vecSubset])
        else:
            h, edges = np.histogram(vecT, bins=np.linspace(0, vecT[-1], num=1000))
            axs[1].stairs(h, edges)
        
        #vecSubset = np.round(np.linspace(1, len(vecT), 1000))
        #axs[1].plot(vecT[vecSubset], vecRate[vecSubset])
       
        if dblMeanRate == 1:
            strLabelY = 'Time-locked activation (a.u.)'
        else:
            strLabelY = 'Spiking rate (Hz)'
        axs[1].set(xlabel='Time (s)', ylabel=strLabelY, title='Peri Event Plot (PEP)')
        

    ## build output
    sMSD = dict()
    sMSD['vecRate'] = vecRate
    sMSD['vecT'] = vecT
    sMSD['vecM'] = vecM
    sMSD['vecScale'] = vecScale
    sMSD['matMSD'] = matMSD
    sMSD['vecV'] = vecV

    return vecRate, sMSD

def getD(dblScale,intS,intN,vecT,vecV):
    # select points within window
    dblT = vecT[intS]
    dblMinEdge = dblT - dblScale/2
    dblMaxEdge = dblT + dblScale/2
    ### intIdxMinT = find(vecT > dblMinEdge,1);
    ### if isempty(intIdxMinT)
    ###     intIdxMinT=1
    ### end
    intIdxMinT = np.where(vecT > dblMinEdge)[0]
    if len(intIdxMinT) > 0:
        intIdxMinT = intIdxMinT[0]
    else:
        intIdxMinT = 0

    ### intIdxMaxT = find(vecT > dblMaxEdge,1);
    ### if isempty(intIdxMaxT)
    ###     intIdxMaxT=intN
    ### end
    intIdxMaxT = np.where(vecT > dblMaxEdge)[0]
    if len(intIdxMaxT) > 0 :
        intIdxMaxT = intIdxMaxT[0]
    else:
        intIdxMaxT = intN - 1

    ### if intIdxMinT == intIdxMaxT && intIdxMinT > 1
    ###     intIdxMinT=intIdxMaxT-1
    ### end
    if intIdxMinT == intIdxMaxT and intIdxMinT > 0 :
        intIdxMinT = intIdxMaxT-1

    ### dbl_dT = max([dblScale (vecT(intIdxMaxT) - vecT(intIdxMinT))]);
    ### dblD = (vecV(intIdxMaxT) - vecV(intIdxMinT))/dbl_dT;
    dbl_dT = max([dblScale, (vecT[intIdxMaxT] - vecT[intIdxMinT])])
    dblD = (vecV[intIdxMaxT] - vecV[intIdxMinT]) / dbl_dT
    return dblD



from setuptools import setup

setup(
    name='zetapy',
    version='3.0.4',
    description='Implementations of the ZETA family of statistical tests.',
    url='https://github.com/JorritMontijn/zetapy',
    author='Jorrit Montijn, Guido Meijer & Alexander Heimel',
    author_email='j.s.montijn@gmail.com',
    license='GNU General Public License',
    packages=['zetapy', 'zetapy.legacy'],
    install_requires=['scipy', 'numpy', 'matplotlib'],

    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



# -*- coding: utf-8 -*-
import numpy as np
import logging
from scipy import stats
from math import pi, sqrt, exp, factorial
from collections.abc import Iterable

# %% calcZetaTwo


def calcZetaTwo(vecSpikeTimes1, arrEventTimes1, vecSpikeTimes2, arrEventTimes2, dblUseMaxDur, intResampNum, boolDirectQuantile):
    """
    Calculates two-sample zeta
    dZETA = calcZetaTwo(vecSpikeTimes1, arrEventTimes1, vecSpikeTimes2,
                        vecEventStarts2, dblUseMaxDur, intResampNum, boolDirectQuantile)
    dZETA has entries:
        vecSpikeT, vecRealDeviation, vecRealFrac, vecRealFracLinear, cellRandTime, cellRandDeviation, dblZetaP, dblZETA, intZETAIdx
    """

    # %% pre-allocate output
    vecSpikeT = None
    vecRealDiff = None
    vecRealFrac1 = None
    vecRealFrac2 = None
    cellRandTime = None
    cellRandDiff = None
    dblZetaP = 1.0
    dblZETA = 0.0
    intZETAIdx = None

    dZETA = dict()
    dZETA['vecSpikeT'] = vecSpikeT
    dZETA['vecRealDiff'] = vecRealDiff
    dZETA['vecRealFrac1'] = vecRealFrac1
    dZETA['vecRealFrac2'] = vecRealFrac2
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDiff'] = cellRandDiff
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx

    # %% ensure input is correct
    # assert that arrEventTimes is a 1D array of floats
    assert len(arrEventTimes1.shape) < 3 and len(arrEventTimes2.shape) < 3 \
        and issubclass(arrEventTimes1.dtype.type, np.floating) and issubclass(arrEventTimes2.dtype.type, np.floating), \
        "Input arrEventTimes1 or arrEventTimes2 is not a 1D or 2D float np.array"

    # ensure orientation arrEventTimes1
    if len(arrEventTimes1.shape) > 1:
        if arrEventTimes1.shape[1] < 3:
            pass
        elif arrEventTimes1.shape[0] < 3:
            arrEventTimes1 = arrEventTimes1.T
        else:
            raise Exception(
                "Input error: arrEventTimes1 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes1 = np.reshape(arrEventTimes1, (-1, 1))

    # ensure orientation arrEventTimes2
    if len(arrEventTimes2.shape) > 1:
        if arrEventTimes2.shape[1] < 3:
            pass
        elif arrEventTimes2.shape[0] < 3:
            arrEventTimes2 = arrEventTimes2.T
        else:
            raise Exception(
                "Input error: arrEventTimes2 must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes2 = np.reshape(arrEventTimes2, (-1, 1))

    # %% get spikes per trial
    vecEventT1 = arrEventTimes1[:, 0]
    vecEventT2 = arrEventTimes2[:, 0]
    cellTrialPerSpike1, cellTimePerSpike1 = getSpikesInTrial(vecSpikeTimes1, vecEventT1, dblUseMaxDur)
    cellTrialPerSpike2, cellTimePerSpike2 = getSpikesInTrial(vecSpikeTimes2, vecEventT2, dblUseMaxDur)

    # %% run normal
    # normalize to cumsum(v1)+cumsum(v2) = 1
    # take difference
    # mean-subtract

    # get difference
    vecSpikeT, vecRealDiff, vecRealFrac1, vecThisSpikeTimes1, vecRealFrac2, vecThisSpikeTimes2 = \
        getTempOffsetTwo(cellTimePerSpike1, cellTimePerSpike2, dblUseMaxDur)

    if len(vecRealDiff) < 2:
        return dZETA

    intZETAIdx = np.argmax(np.abs(vecRealDiff))
    dblMaxD = np.abs(vecRealDiff[intZETAIdx])

    # repeat procedure, but swap trials randomly in each resampling
    cellRandTime = [None] * intResampNum
    cellRandDiff = [None] * intResampNum
    vecMaxRandD = np.empty((intResampNum, 1))
    vecMaxRandD.fill(np.nan)

    cellAggregateTrials = cellTimePerSpike1 + cellTimePerSpike2
    intTrials1 = len(cellTimePerSpike1)
    intTrials2 = len(cellTimePerSpike2)
    intTotTrials = intTrials1+intTrials2

    # %% run bootstraps; try parallel, otherwise run normal loop
    # repeat procedure, but swap trials randomly in each resampling
    for intResampling in range(intResampNum):
        # %% get random subsample
        # if cond1 has 10 trials, and cond2 has 100, then:
        # for shuffle of cond1: take 10 trials from set of 110
        # for shuffle of cond2: take 100 trials from set of 110
        vecUseRand1 = np.random.randint(intTotTrials, size=intTrials1)
        vecUseRand2 = np.random.randint(intTotTrials, size=intTrials2)

        cellTimePerSpike1_Rand = [cellAggregateTrials[i] for i in vecUseRand1]
        cellTimePerSpike2_Rand = [cellAggregateTrials[j] for j in vecUseRand2]

        if np.sum([len(xi) for xi in cellTimePerSpike1_Rand]) == 0 and np.sum([len(yi) for yi in cellTimePerSpike2_Rand]) == 0:
            dblAddVal = None
        else:
            # get difference
            vecRandT, vecRandDiff, vecRandFrac1, vecThisSpikeTimes1, vecRandFrac2, vecThisSpikeTimes2 = \
                getTempOffsetTwo(cellTimePerSpike1_Rand, cellTimePerSpike2_Rand, dblUseMaxDur)

            # assign data
            cellRandTime[intResampling] = vecRandT
            cellRandDiff[intResampling] = vecRandDiff
            dblAddVal = np.max(np.abs(vecRandDiff))

        # assign read-out
        if dblAddVal is None or dblAddVal == 0:
            dblAddVal = dblMaxD
        vecMaxRandD[intResampling] = dblAddVal

    # %% calculate significance
    dblZetaP, dblZETA = getZetaP(dblMaxD, vecMaxRandD, boolDirectQuantile)

    # %% return
    dZETA = dict()
    dZETA['vecSpikeT'] = vecSpikeT
    dZETA['vecRealDiff'] = vecRealDiff
    dZETA['vecRealFrac1'] = vecRealFrac1
    dZETA['vecRealFrac2'] = vecRealFrac2
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDiff'] = cellRandDiff
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx
    return dZETA

# %%


def calcZetaOne(vecSpikeTimes, arrEventTimes, dblUseMaxDur, intResampNum, boolDirectQuantile, dblJitterSize, boolStitch, boolParallel):
    """
    Calculates neuronal responsiveness index zeta
    dZETA = calcZetaOne(vecSpikeTimes, vecEventStarts, dblUseMaxDur, intResampNum, boolDirectQuantile, dblJitterSize, boolStitch,boolParallel, intUseJitterDistro)
    dZETA has entries:
        vecSpikeT, vecRealDeviation, vecRealFrac, vecRealFracLinear, cellRandTime, cellRandDeviation, dblZetaP, dblZETA, intZETAIdx
    """

    # %% pre-allocate output
    vecSpikeT = None
    vecRealDeviation = None
    vecRealFrac = None
    vecRealFracLinear = None
    cellRandTime = None
    cellRandDeviation = None
    dblZetaP = 1.0
    dblZETA = 0.0
    intZETAIdx = None

    dZETA = dict()
    dZETA['vecSpikeT'] = vecSpikeT
    dZETA['vecRealDeviation'] = vecRealDeviation
    dZETA['vecRealFrac'] = vecRealFrac
    dZETA['vecRealFracLinear'] = vecRealFracLinear
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDeviation'] = cellRandDeviation
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx

    # %% prep parallel processing
    # to do

    # %% reduce spikes
    # ensure orientation and assert that arrEventTimes is a 1D array of floats
    assert len(arrEventTimes.shape) < 3 and issubclass(
        arrEventTimes.dtype.type, np.floating), "Input arrEventTimes is not a 1D or 2D float np.array"
    if len(arrEventTimes.shape) > 1:
        if arrEventTimes.shape[1] < 3:
            pass
        elif arrEventTimes.shape[0] < 3:
            arrEventTimes = arrEventTimes.T
        else:
            raise Exception(
                "Input error: arrEventTimes must be T-by-1 or T-by-2; with T being the number of trials/stimuli/events")
    else:
        # turn into T-by-1 array
        arrEventTimes = np.reshape(arrEventTimes, (-1, 1))
    # define event starts
    vecEventT = arrEventTimes[:, 0]

    dblMinPreEventT = np.min(vecEventT)-dblUseMaxDur*5*dblJitterSize
    dblStartT = max([vecSpikeTimes[0], dblMinPreEventT])
    dblStopT = max(vecEventT)+dblUseMaxDur*5*dblJitterSize
    vecSpikeTimes = vecSpikeTimes[np.logical_and(vecSpikeTimes >= dblStartT, vecSpikeTimes <= dblStopT)]

    if vecSpikeTimes.size < 3:
        logging.warning(
            "calcZetaOne:vecSpikeTimes: too few spikes around events to calculate zeta")
        return dZETA

    # %% build pseudo data, stitching stimulus periods
    if boolStitch:
        vecPseudoSpikeTimes, vecPseudoEventT = getPseudoSpikeVectors(vecSpikeTimes, vecEventT, dblUseMaxDur)
    else:
        vecPseudoSpikeTimes = vecSpikeTimes
        vecPseudoEventT = vecEventT

    # %% run normal
    # get data
    vecRealDeviation, vecRealFrac, vecRealFracLinear, vecSpikeT = getTempOffsetOne(
        vecPseudoSpikeTimes, vecPseudoEventT, dblUseMaxDur)

    if vecRealDeviation.size < 3:
        logging.warning(
            "calcZetaOne:vecRealDeviation: too few spikes around events to calculate zeta")
        return dZETA

    vecRealDeviation = vecRealDeviation - np.mean(vecRealDeviation)
    intZETAIdx = np.argmax(np.abs(vecRealDeviation))
    dblMaxD = np.abs(vecRealDeviation[intZETAIdx])

    # %% create random jitters
    # run pre-set number of iterations
    cellRandTime = []
    cellRandDeviation = []
    vecMaxRandD = np.empty((intResampNum, 1))
    vecMaxRandD.fill(np.nan)

    vecStartOnly = np.reshape(vecPseudoEventT, (-1, 1))
    intTrials = vecStartOnly.size
    matJitterPerTrial = np.empty((intTrials, intResampNum))
    matJitterPerTrial.fill(np.nan)

    # uniform jitters between dblJitterSize*[-tau, +tau]
    for intResampling in range(intResampNum):
        matJitterPerTrial[:, intResampling] = dblJitterSize*dblUseMaxDur * \
            ((np.random.rand(vecStartOnly.shape[0]) - 0.5) * 2)

    # %% this part is only to check if matlab and python give the same exact results
    # unfortunately matlab's randperm() and numpy's np.random.permutation give different outputs even with
    # identical seeds and identical random number generators, so I've had to load in a table of random values here...
    boolTest = False
    if boolTest:
        from scipy.io import loadmat
        print('Loading deterministic jitter data for comparison with matlab')
        logging.warning(
            "calcZetaOne:debugMode: set boolTest to False to suppress this warning")
        dLoad = loadmat('matJitterPerTrial.mat')
        matJitterPerTrial = dLoad['matJitterPerTrial']

        # reset rng
        np.random.seed(1)

    # %% run resamplings
    for intResampling in range(intResampNum):
        # get random subsample
        vecStimUseOnTime = vecStartOnly[:, 0] + matJitterPerTrial[:, intResampling].T

        # get temp offset
        vecRandDiff, vecThisSpikeFracs, vecThisFracLinear, vecThisSpikeTimes = getTempOffsetOne(
            vecPseudoSpikeTimes, vecStimUseOnTime, dblUseMaxDur)

        # assign data
        cellRandTime.append(vecThisSpikeTimes)
        cellRandDeviation.append(vecRandDiff - np.mean(vecRandDiff))
        vecMaxRandD[intResampling] = np.max(np.abs(cellRandDeviation[intResampling]))

    # %% calculate significance
    dblZetaP, dblZETA = getZetaP(dblMaxD, vecMaxRandD, boolDirectQuantile)

    # %% assign output
    dZETA = dict()
    dZETA['vecSpikeT'] = vecSpikeT
    dZETA['vecRealDeviation'] = vecRealDeviation
    dZETA['vecRealFrac'] = vecRealFrac
    dZETA['vecRealFracLinear'] = vecRealFracLinear
    dZETA['cellRandTime'] = cellRandTime
    dZETA['cellRandDeviation'] = cellRandDeviation
    dZETA['dblZetaP'] = dblZetaP
    dZETA['dblZETA'] = dblZETA
    dZETA['intZETAIdx'] = intZETAIdx
    return dZETA

# %%


def getTempOffsetTwo(cellTimePerSpike1, cellTimePerSpike2, dblUseMaxDur):
    '''
    vecSpikeT, vecThisDiff, vecThisFrac1, vecThisSpikeTimes1, vecThisFrac2, vecThisSpikeTimes2 = 
        getTempOffsetTwo(cellTimePerSpike1,cellTimePerSpike2,dblUseMaxDur)
    '''

    # introduce minimum jitter to identical spikes
    vecSpikes1 = flatten(cellTimePerSpike1)
    vecSpikes2 = flatten(cellTimePerSpike2)
    
    vecThisSpikeTimes1 = getUniqueSpikes(np.sort(vecSpikes1))
    vecThisSpikeTimes2 = getUniqueSpikes(np.sort(vecSpikes2))

    # ref time
    vecSpikeT = np.sort(np.concatenate((np.zeros(1), vecThisSpikeTimes1,
                        vecThisSpikeTimes2, np.array([dblUseMaxDur])), axis=0))

    # cond1 goes to S1_n/T1_n; cond2 goes to S2_n/T2_n
    intSp1 = len(vecThisSpikeTimes1)
    intSp2 = len(vecThisSpikeTimes2)
    intT1 = len(cellTimePerSpike1)
    intT2 = len(cellTimePerSpike2)

    # spike fraction #1
    vecUniqueSpikeFracs1 = np.linspace(1, vecThisSpikeTimes1.size, vecThisSpikeTimes1.size)/intT1
    vecSpikes1 = np.concatenate((np.zeros(1), vecThisSpikeTimes1, np.array([dblUseMaxDur])), axis=0)
    vecFracs1 = np.concatenate((np.zeros(1), vecUniqueSpikeFracs1, np.array([intSp1/intT1])), axis=0)
    vecThisFrac1 = np.interp(vecSpikeT,
                             vecSpikes1,
                             vecFracs1,
                             1/intT1, intSp1/intT1)

    # spike fraction #2
    vecUniqueSpikeFracs2 = np.linspace(1, vecThisSpikeTimes2.size, vecThisSpikeTimes2.size)/intT2
    vecSpikes2 = np.concatenate((np.zeros(1), vecThisSpikeTimes2, np.array([dblUseMaxDur])), axis=0)
    vecFracs2 = np.concatenate((np.zeros(1), vecUniqueSpikeFracs2, np.array([intSp2/intT2])), axis=0)
    vecThisFrac2 = np.interp(vecSpikeT,
                             vecSpikes2,
                             vecFracs2,
                             1/intT2, intSp2/intT2)

    # take difference
    vecDeviation = vecThisFrac1 - vecThisFrac2

    # mean-subtract?
    vecThisDiff = vecDeviation - np.mean(vecDeviation)

    return vecSpikeT, vecThisDiff, vecThisFrac1, vecThisSpikeTimes1, vecThisFrac2, vecThisSpikeTimes2

# %%


def getSpikesInTrial(vecSpikes, vecTrialStarts, dblMaxDur):
    '''
    getSpikesInTrial Retrieves spiking times per trial
    syntax: cellTrialPerSpike,cellTimePerSpike = getSpikesInTrial(vecSpikes,vecTrialStarts,dblMaxDur)
    input:
        - vecSpikes; spike times (s)
        - vecTrialStarts: trial start times (s)
        - dblTrialDur: trial duration (s)
    returns:
        - cellTrialPerSpike
        - cellTimePerSpike
    '''

    # loop
    intTrials = len(vecTrialStarts)
    cellTrialPerSpike = []
    cellTimePerSpike = []
    for intTrial, dblStartT in enumerate(vecTrialStarts):
        # get spikes
        vecTheseSpikes = vecSpikes[np.logical_and(vecSpikes >= dblStartT, vecSpikes < (dblStartT + dblMaxDur))] - dblStartT

        # assign
        cellTrialPerSpike.append(intTrial*np.ones(len(vecTheseSpikes)))
        cellTimePerSpike.append(vecTheseSpikes)

    return cellTrialPerSpike, cellTimePerSpike

# %%


def getZetaP(arrMaxD, vecMaxRandD, boolDirectQuantile):
    # %% calculate significance
    # find highest peak and retrieve value
    vecMaxRandD = np.sort(np.unique(vecMaxRandD), axis=0)
    if not isinstance(arrMaxD, Iterable):
        arrMaxD = np.array([arrMaxD])

    if boolDirectQuantile:
        # calculate statistical significance using empirical quantiles
        # define p-value
        arrZetaP = np.empty(arrMaxD.size)
        arrZetaP.fill(np.nan)
        for i, d in enumerate(arrMaxD):
            if d < np.min(vecMaxRandD) or np.isnan(d):
                dblValue = 0
            elif d > np.max(vecMaxRandD) or np.isinf(d):
                dblValue = vecMaxRandD.size
            else:
                dblValue = np.interp(
                    d, vecMaxRandD, np.arange(0, vecMaxRandD.size)+1)

            arrZetaP[i] = 1 - (dblValue/(1+vecMaxRandD.size))

        # transform to output z-score
        arrZETA = -stats.norm.ppf(arrZetaP/2)
    else:
        # calculate statistical significance using Gumbel distribution
        arrZetaP, arrZETA = getGumbel(
            np.mean(vecMaxRandD), np.var(vecMaxRandD, ddof=1), arrMaxD)  # default ddof for numpy var() is incorrect

    # return
    if arrZetaP.size == 1:
        arrZetaP = arrZetaP[0]
    if arrZETA.size == 1:
        arrZETA = arrZETA[0]
    return arrZetaP, arrZETA

# %%


def getGumbel(dblE, dblV, arrX):
    """"Calculate p-value and z-score for maximum value of N samples drawn from Gaussian
           dblP,dblZ = getGumbel(dblE,dblV,arrX)

                input:
                - dblE: mean of distribution of maximum values
                - dblV: variance of distribution of maximum values
                - arrX: maximum value to express in quantiles of Gumbel

                output:
                - arrP; p-value for dblX (chance that sample originates from distribution given by dblE/dblV)
                - arrZ; z-score corresponding to P

        Version history:
        1.0 - June 17, 2020
        Created by Jorrit Montijn, translated by Alexander Heimel
        3.0 - August 17 2023
        New translation to Python by Jorrit Montijn: Now supports array input of arrX

        Sources:
        Baglivo (2005)
        Elfving (1947), https://doi.org/10.1093/biomet/34.1-2.111
        Royston (1982), DOI: 10.2307/2347982
        https://stats.stackexchange.com/questions/394960/variance-of-normal-order-statistics
        https://stats.stackexchange.com/questions/9001/approximate-order-statistics-for-normal-random-variables
        https://en.wikipedia.org/wiki/Extreme_value_theory
        https://en.wikipedia.org/wiki/Gumbel_distribution
    """

    # %% define constants
    # define Euler-Mascheroni constant
    dblEulerMascheroni = 0.5772156649015328606065120900824  # vpa(eulergamma)

    # %% define Gumbel parameters from mean and variance
    # derive beta parameter from variance
    dblBeta = (sqrt(6)*sqrt(dblV))/(pi)

    # derive mode from mean, beta and E-M constant
    dblMode = dblE - dblBeta*dblEulerMascheroni

    # define Gumbel cdf
    def fGumbelCDF(x): return np.exp(-np.exp(-((x-dblMode) / dblBeta)))

    # %% calculate output variables
    # calculate cum dens at X
    arrGumbelCDF = fGumbelCDF(arrX)

    # define p-value
    arrP = 1-arrGumbelCDF

    # transform to output z-score
    arrZ = -stats.norm.ppf(np.divide(arrP, 2))

    # approximation for large X
    for i, dblZ in enumerate(arrZ):
        if np.isinf(dblZ):
            arrP[i] = exp(dblMode-arrX[i] / dblBeta)
            arrZ[i] = -stats.norm.ppf(arrP[i]/2)

    # return
    return arrP, arrZ

# %%


def getTempOffsetOne(vecSpikeTimes, vecEventTimes, dblUseMaxDur):
    # %% get temp diff vector
    # pre-allocate
    vecSpikesInTrial = getSpikeT(vecSpikeTimes, vecEventTimes, dblUseMaxDur)

    # introduce minimum jitter to identical spikes
    vecThisSpikeTimes = getUniqueSpikes(vecSpikesInTrial)

    # turn into fractions
    vecThisSpikeFracs = np.linspace(
        1/vecThisSpikeTimes.size, 1, vecThisSpikeTimes.size)

    # get linear fractions
    vecThisFracLinear = vecThisSpikeTimes/dblUseMaxDur

    # calc difference
    vecThisDeviation = vecThisSpikeFracs - vecThisFracLinear
    vecThisDeviation = vecThisDeviation - np.mean(vecThisDeviation)

    return vecThisDeviation, vecThisSpikeFracs, vecThisFracLinear, vecThisSpikeTimes

# %%


def getUniqueSpikes(vecSpikesInTrial):
    # introduce random minimum jitter to identical spikes
    dblMinOffset = np.finfo(vecSpikesInTrial.dtype.type).eps
    vecOffsets = np.arange(-dblMinOffset*5, dblMinOffset*5, dblMinOffset)
    vecUniqueSpikes, vecIdx = np.unique(vecSpikesInTrial, return_index=True)
    while vecUniqueSpikes.shape[0] != vecSpikesInTrial.shape[0]:
        indDuplicates = ~np.isin(np.arange(vecSpikesInTrial.shape[0]), vecIdx)
        vecRandomOffsets = np.random.choice(vecOffsets, np.sum(indDuplicates))
        vecSpikesInTrial[indDuplicates] += vecRandomOffsets
        vecUniqueSpikes, vecIdx = np.unique(vecSpikesInTrial, return_index=True)
    return vecSpikesInTrial

# %%


def getSpikeT(vecSpikeTimes, vecEventTimes, dblUseMaxDur):
    # %% turn spike times relative to recording start into times relative to trial start

    # pre-allocate
    vecSpikesInTrial = np.empty((vecSpikeTimes.size*2))
    vecSpikesInTrial.fill(np.nan)
    intIdx = 0

    # go through trials to build spike time vector
    for dblStartT in vecEventTimes:
        # get times
        dblStopT = dblStartT + dblUseMaxDur

        # build trial assignment
        vecTempSpikes = vecSpikeTimes[np.logical_and(vecSpikeTimes < dblStopT, vecSpikeTimes > dblStartT)] - dblStartT
        intTempSpikeNr = vecTempSpikes.size
        vecAssignIdx = [i for i in range(intIdx, intIdx+intTempSpikeNr)]
        if len(vecAssignIdx) > 0 and vecAssignIdx[-1] >= vecSpikesInTrial.size:
            vecSpikesInTrial = np.resize(vecSpikesInTrial, vecSpikesInTrial.size*2)
        vecSpikesInTrial[vecAssignIdx] = vecTempSpikes
        intIdx = intIdx + intTempSpikeNr

    # remove trailing nan entries
    vecSpikesInTrial = vecSpikesInTrial[:intIdx]

    # sort spikes in window and add start/end entries
    vecSpikesInTrial = np.concatenate((np.zeros(1), np.sort(vecSpikesInTrial, axis=0), np.array([dblUseMaxDur])))

    return vecSpikesInTrial

# %%


def getPseudoSpikeVectors(vecSpikeTimes, vecEventTimes, dblWindowDur, boolDiscardEdges=False):
    # %% prep
    # ensure sorting and alignment
    vecSpikeTimes = np.sort(np.reshape(vecSpikeTimes, (-1, 1)), axis=0)
    vecEventTimes = np.sort(np.reshape(vecEventTimes, (-1, 1)), axis=0)

    # %% pre-allocate
    intSamples = vecSpikeTimes.size
    intTrials = vecEventTimes.size
    dblMedianDur = np.median(np.diff(vecSpikeTimes, axis=0))
    cellPseudoSpikeT = []
    vecPseudoEventT = np.empty((intTrials, 1))
    vecPseudoEventT.fill(np.nan)
    dblPseudoEventT = 0.0
    intLastUsedSample = 0
    intFirstSample = None

    # run
    for intTrial, dblEventT in enumerate(vecEventTimes):
        # get eligible samples
        intStartSample = findfirst(vecSpikeTimes >= dblEventT)
        intEndSample = findfirst(vecSpikeTimes > (dblEventT+dblWindowDur))

        if intStartSample is not None and intEndSample is not None and intStartSample > intEndSample:
            intEndSample = None
            intStartSample = None

        if intEndSample is None:
            intEndSample = len(vecSpikeTimes)

        if intStartSample is None or intEndSample is None:
            vecUseSamples = np.empty(0, dtype=int)
        else:
            intEndSample = intEndSample - 1
            vecEligibleSamples = np.arange(intStartSample, intEndSample+1)
            indUseSamples = np.logical_and(vecEligibleSamples >= 0, vecEligibleSamples < intSamples)
            vecUseSamples = vecEligibleSamples[indUseSamples]

        # check if beginning or end
        if vecUseSamples.size > 0:
            if intTrial == 0 and not boolDiscardEdges:
                vecUseSamples = np.arange(0, vecUseSamples[-1]+1)
            elif intTrial == (intTrials-1) and not boolDiscardEdges:
                vecUseSamples = np.arange(vecUseSamples[0], intSamples)

        # add spikes
        if vecUseSamples.size > 0:
            vecAddT = vecSpikeTimes[vecUseSamples]
            indOverlap = vecUseSamples <= intLastUsedSample

        # get event t
        if intTrial == 0:
            dblPseudoEventT = 0.0
        else:
            if intTrial > 0 and dblWindowDur > (dblEventT - vecEventTimes[intTrial-1]):
                # remove spikes from overlapping epochs
                if vecUseSamples.size > 0:
                    vecUseSamples = vecUseSamples[~indOverlap]
                    vecAddT = vecSpikeTimes[vecUseSamples]

                dblPseudoEventT = dblPseudoEventT + dblEventT - vecEventTimes[intTrial-1]
            else:
                dblPseudoEventT = dblPseudoEventT + dblWindowDur

        # %% make local pseudo event time
        if vecUseSamples.size == 0:
            vecLocalPseudoT = np.empty(0)
        else:
            intLastUsedSample = vecUseSamples[-1]
            vecLocalPseudoT = vecAddT - dblEventT + dblPseudoEventT

        if intFirstSample is None and vecUseSamples.size > 0:
            intFirstSample = vecUseSamples[0]
            dblPseudoT0 = dblPseudoEventT

        # assign data for this trial
        cellPseudoSpikeT.append(vecLocalPseudoT)
        vecPseudoEventT[intTrial] = dblPseudoEventT

    # %% add beginning
    if not boolDiscardEdges and intFirstSample is not None and intFirstSample > 0:
        dblStepBegin = vecSpikeTimes[intFirstSample] - vecSpikeTimes[intFirstSample-1]
        vecSampAddBeginning = np.arange(0, intFirstSample)
        vecAddBeginningSpikes = vecSpikeTimes[vecSampAddBeginning] - vecSpikeTimes[vecSampAddBeginning[0]] \
            + dblPseudoT0 - dblStepBegin - \
            np.ptp(vecSpikeTimes[vecSampAddBeginning]
                   )  # make local to first spike in array, then preceding pseudo event t0
        cellPseudoSpikeT.append(vecAddBeginningSpikes)

    # %% add end
    intTn = vecSpikeTimes.size
    intLastUsedSample = findfirst(vecSpikeTimes > (vecEventTimes[-1]+dblWindowDur))
    if not boolDiscardEdges and intLastUsedSample is not None and (intTn-1) > intLastUsedSample:
        vecSampAddEnd = np.arange(intLastUsedSample, intTn)
        vecAddEndSpikes = vecSpikeTimes[vecSampAddEnd] - dblEventT + dblPseudoEventT + dblWindowDur
        cellPseudoSpikeT.append(vecAddEndSpikes)

    # %% recombine into vector
    vecPseudoSpikeTimes = np.array(sorted(flatten(cellPseudoSpikeT)))
    return vecPseudoSpikeTimes, vecPseudoEventT

# %%


def findfirst(indArray):
    vecStartSamples = np.where(indArray)[0]
    if vecStartSamples.size == 0:
        intStartSample = None
    else:
        intStartSample = vecStartSamples[0]
    return intStartSample

# %%


def flatten(l):
    g = genFlatten(l)
    x = []
    for i,v in enumerate(g):
        x.append(v)
    return x

def genFlatten(l):
    for el in l:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 18 11:21:03 2023

@author: Jorrit
"""

import logging
from scipy.stats import norm
from scipy.signal import convolve2d
import numpy as np
from zetapy.dependencies import (flatten, findfirst)
from scipy import stats, interpolate, signal


def getMultiScaleDeriv(vecT, vecV,
                       dblSmoothSd=2.0, dblMinScale=None, dblBase=1.5, dblMeanRate=1.0, dblUseMaxDur=None, boolParallel=False):
    """"Returns multi-scale derivative of the deviation vector; i.e., the ZETA-derived instantaneous firing rate
       [vecRate,dMSD] = getMultiScaleDeriv(vecT, vecV,
                              dblSmoothSd=2.0, dblMinScale=None, dblBase=1.5, dblMeanRate=1.0, dblUseMaxDur=None, boolParallel=False)

        Required input:
            - vecT [N x 1]: timestamps (e.g., spike times)
            - vecV [N x 1]: values (e.g., deviation vector)

        Optional inputs:
            - dblSmoothSd: Gaussian SD of smoothing kernel (in # of samples) [default: 0.0]
            - dblMinScale: minimum derivative scale [default: round(log(1/1000) / log(dblBase))]
            - dblBase: base for exponential scale step size [default: 1.5]
            - dblMeanRate: mean spiking rate to normalize vecRate [default: None; returns unscaled derivatives]
            - dblUseMaxDur: trial duration to normalize vecRate [default: np.ptp(vecT)]

        Outputs:
            - vecRate; Instantaneous spiking rate
            - dMSD; dictionary with entries:
                - vecRate; instantaneous spiking rates (same as first output)
                - vecT; time-points corresponding to vecRate (same as input vecT)
                - vecM; Mean of multi-scale derivatives (same as vecRate, but before rescaling)
                - vecScale; timescales used to calculate derivatives
                - matMSD; multi-scale derivatives matrix
                - vecV; values on which vecRate is calculated (same as input vecV)
                - dblSmoothSd; smoothing strength (same as input dblSmoothSd)
                - dblMeanRate; mean rate used for rescaling (same as input dblMeanRate)

        Version history:
        2.0 - 2023-08-22
            Created by Jorrit Montijn - updated translation to Python.
    """

    # %% check inputs
    # trial dur
    dblRange = np.ptp(vecT)
    if dblUseMaxDur is None:
        dblUseMaxDur = dblRange

    # min scale
    if dblMinScale is None:
        dblMinScale = round(np.log(1/1000) / np.log(dblBase))

    # parallel processing: to do
    boolUseParallel = False

    # flatten and reorder vecT
    vecT = vecT.flatten()
    vecV = vecV.flatten()
    vecReorder = np.argsort(vecT, axis=0)
    vecT = vecT[vecReorder]
    vecV = vecV[vecReorder]
    indKeep = ~np.logical_or(vecT == 0, vecT == dblUseMaxDur)
    vecT = vecT[indKeep]
    vecV = vecV[indKeep]

    # %% get multi-scale derivative
    dblMaxScale = np.log(dblRange/10) / np.log(dblBase)
    vecExp = np.arange(dblMinScale, dblMaxScale)
    vecScale = dblBase**vecExp
    intScaleNum = len(vecScale)
    intN = len(vecT)
    matMSD = np.zeros((intN, intScaleNum))

    if boolUseParallel and False:
        # not implemented yet
        pass
    else:
        for intScaleIdx, dblScale in enumerate(vecScale):
            # run through all points
            matMSD[:, intScaleIdx] = calcSingleMSD(dblScale, vecT, vecV)

    # %% smoothing
    if dblSmoothSd > 0:
        intSmoothRange = 2*np.ceil(dblSmoothSd).astype(int)
        vecFilt = norm.pdf(range(-intSmoothRange, intSmoothRange+1), 0, dblSmoothSd)
        vecFilt = vecFilt / sum(vecFilt)

        # pad array
        intPadSize = np.floor(len(vecFilt)/2).astype(int)
        matMSD = np.pad(matMSD, ((intPadSize, intPadSize), (0, 0)), 'edge')

        # filter
        matMSD = convolve2d(matMSD, np.reshape(vecFilt, (-1, 1)), 'valid')

    # mean
    vecM = np.mean(matMSD, axis=1)

    # weighted average of vecM by inter-spike intervals
    dblMeanM = (1.0/dblUseMaxDur) * sum(((vecM[:-1] + vecM[1:])/2.0) * np.diff(vecT))

    # rescale to real firing rates
    vecRate = dblMeanRate * ((vecM + 1.0/dblUseMaxDur)/(dblMeanM + 1.0/dblUseMaxDur))

    # %% build output
    dMSD = dict()
    dMSD['vecRate'] = vecRate
    dMSD['vecT'] = vecT
    dMSD['vecM'] = vecM
    dMSD['vecScale'] = vecScale
    dMSD['matMSD'] = matMSD
    dMSD['vecV'] = vecV
    dMSD['dblSmoothSd'] = dblSmoothSd
    dMSD['dblMeanRate'] = dblMeanRate

    return vecRate, dMSD

# %%


def calcSingleMSD(dblScale, vecT, vecV):
    intN = vecT.size
    vecMSD = np.zeros((intN,))

    # run through all points
    for intS, dblT in enumerate(vecT):
        # select points within window

        dblMinEdge = dblT - dblScale/2
        dblMaxEdge = dblT + dblScale/2
        intIdxMinT = findfirst(vecT > dblMinEdge)
        if intIdxMinT is None:
            intIdxMinT = 0

        intIdxMaxT = findfirst(vecT > dblMaxEdge)
        if intIdxMaxT is None:
            intIdxMaxT = intN - 1
        else:
            intIdxMaxT = intIdxMaxT - 1

        if (intIdxMinT > intIdxMaxT):
            dblD = 0
        else:
            if (intIdxMinT == intIdxMaxT) and (intIdxMinT > 0) and (intIdxMinT < (intN-1)):
                intIdxMaxT = intIdxMinT + 1
                intIdxMinT = intIdxMinT - 1

            dbl_dT = np.max([dblScale, (vecT[intIdxMaxT] - vecT[intIdxMinT])])
            dblD = (vecV[intIdxMaxT] - vecV[intIdxMinT]) / dbl_dT

        # select points within window
        vecMSD[intS] = dblD

    # return single msd vector
    return vecMSD

# %%


def getPeak(vecData, vecT, tplRestrictRange=(-np.inf, np.inf), intSwitchZ=1):
    """Returns highest peak time, width, and location. Syntax:
        getPeak(vecData, vecT, tplRestrictRange=(-np.inf, np.inf), intSwitchZ=1)

    Required input:
        - vecData [N x 1]: values

    Optional inputs:
        - vecT [N x 1]: timestamps corresponding to vecData (default: [1:N])
        - vecRestrictRange: restrict peak to lie within vecRestrictRange(1) and vecRestrictRange(end)

    Output:
        dPeak, dict with entries:
        - dblLatencyPeak: time of peak
        - dblPeakValue: value of peak (rate)
        - dblPeakWidth: width of peak
        - vecPeakStartStop: start/stop times of peak
        - intPeakLoc: index of peak
        - vecPeakStartStopIdx: start/stop indices of peak

    Version history:
    1.0 - June 19, 2020, Created by Jorrit Montijn, Translated to python by Alexander Heimel
    """

    # check inputs
    if len(vecT) == 0:
        vecT = np.arange(len(vecData))

    # z-score
    if intSwitchZ == 1:
        vecDataZ = stats.zscore(vecData)
    elif intSwitchZ == 2:
        dblMu = np.mean(vecData[vecT < 0.02])
        vecDataZ = (vecData - dblMu) / np.std(vecData)
    else:
        vecDataZ = vecData

    # get most prominent POSITIVE peak times
    vecLocsPos, peakProps = signal.find_peaks(vecDataZ, threshold=0, prominence=-np.inf)
    vecValsPos = vecDataZ[vecLocsPos]
    vecPromsPos = peakProps['prominences']

    # remove peaks outside window
    indKeepPeaks = (vecT[vecLocsPos] >= tplRestrictRange[0]) & (vecT[vecLocsPos] <= tplRestrictRange[1])

    if np.sum(indKeepPeaks) == 0:
        dblMaxPosVal = None
    else:
        # select peak
        vecValsPos = vecValsPos[indKeepPeaks]
        vecLocsPos = vecLocsPos[indKeepPeaks]
        vecPromsPos = vecPromsPos[indKeepPeaks]
        intPosIdx = np.argmax(vecValsPos)
        dblMaxPosVal = vecValsPos[intPosIdx]

    # get most prominent NEGATIVE peak times
    vecLocsNeg, peakProps = signal.find_peaks(-vecDataZ, threshold=0, prominence=-np.inf)
    vecValsNeg = -vecDataZ[vecLocsNeg]
    vecPromsNeg = peakProps['prominences']

    # remove peaks outside window
    indKeepPeaks = (vecT[vecLocsNeg] >= tplRestrictRange[0]) & (vecT[vecLocsNeg] <= tplRestrictRange[1])

    if np.sum(indKeepPeaks) == 0:
        dblMaxNegVal = None
    else:
        # select peak
        vecValsNeg = vecValsNeg[indKeepPeaks]
        vecLocsNeg = vecLocsNeg[indKeepPeaks]
        vecPromsNeg = vecPromsNeg[indKeepPeaks]
        intNegIdx = np.argmax(vecValsNeg)
        dblMaxNegVal = vecValsNeg[intNegIdx]

    if dblMaxPosVal is None and dblMaxNegVal is None:
        indPeakMembers = None
    elif ((dblMaxPosVal is not None and dblMaxNegVal is None)
          or (dblMaxPosVal is not None and (np.abs(dblMaxPosVal) >= np.abs(dblMaxNegVal)))):
        intIdx = intPosIdx
        intPeakLoc = vecLocsPos[intIdx]
        dblPeakProm = vecPromsPos[intIdx]
        dblCutOff = vecDataZ[intPeakLoc] - dblPeakProm / 2
        indPeakMembers = (vecDataZ > dblCutOff)
    elif ((dblMaxPosVal is None and dblMaxNegVal is not None)
          or (dblMaxNegVal is not None and (np.abs(dblMaxPosVal) < np.abs(dblMaxNegVal)))):
        intIdx = intNegIdx
        intPeakLoc = vecLocsNeg[intIdx]
        dblPeakProm = vecPromsNeg[intIdx]
        dblCutOff = vecDataZ[intPeakLoc] + dblPeakProm / 2
        indPeakMembers = (vecDataZ < dblCutOff)

    if indPeakMembers is not None:
        # get potential starts/stops
        # vecPeakStarts = find(diff(indPeakMembers)==1);
        vecPeakStarts = np.where(np.diff([float(f) for f in indPeakMembers]) == 1)[0]
        # vecPeakStops = find(diff(indPeakMembers)==-1);
        vecPeakStops = np.where(np.diff([float(f) for f in indPeakMembers]) == -1)[0]
        if indPeakMembers[0]:
            # vecPeakStarts = [1 vecPeakStarts(:)'];
            vecPeakStarts = [0] + vecPeakStarts
        # if indPeakMembers(end) == 1
        if indPeakMembers[-1]:
            # vecPeakStops = [vecPeakStops(:)' numel(indPeakMembers)];
            vecPeakStops = vecPeakStops + [len(indPeakMembers)-1]

        # find closest points
        # intPeakStart = intPeakLoc-min(intPeakLoc - vecPeakStarts(vecPeakStarts<intPeakLoc));

        intPeakStart = intPeakLoc - np.min(intPeakLoc - vecPeakStarts[vecPeakStarts < intPeakLoc])
        intPeakStop = intPeakLoc + np.min(vecPeakStops[vecPeakStops >= intPeakLoc] - intPeakLoc)
        dblPeakStartT = vecT[intPeakStart]
        if intPeakStop >= vecT.shape[0]:
            intPeakStop = vecT.shape[0] - 1
        dblPeakStopT = vecT[intPeakStop]
        # assign peak data
        dblPeakValue = vecData[intPeakLoc]
        dblLatencyPeak = vecT[intPeakLoc]
        dblPeakWidth = dblPeakStopT - dblPeakStartT
        vecPeakStartStop = [dblPeakStartT, dblPeakStopT]
        vecPeakStartStopIdx = [intPeakStart, intPeakStop]
    else:
        # assign placeholder peak data
        dblPeakValue = np.nan
        dblLatencyPeak = np.nan
        dblPeakWidth = np.nan
        vecPeakStartStop = [np.nan, np.nan]
        intPeakLoc = None
        vecPeakStartStopIdx = [None, None]

    dPeak = dict()
    dPeak['dblLatencyPeak'] = dblLatencyPeak
    dPeak['dblPeakValue'] = dblPeakValue
    dPeak['dblPeakWidth'] = dblPeakWidth
    dPeak['vecPeakStartStop'] = vecPeakStartStop
    dPeak['intPeakLoc'] = intPeakLoc
    dPeak['vecPeakStartStopIdx'] = vecPeakStartStopIdx
    return dPeak


# %%
def getOnset(vecData, vecT, dblLatencyPeak=None, tplRestrictRange=None):
    """Returns peak onset. Syntax:
        dOnset = getOnset(vecData,vecT,dblLatencyPeak=None,tplRestrictRange=None)

    Required input:
        - vecData [N x 1]: values
        - vecT [N x 1]: timestamps corresponding to vecData

    Optional inputs:
        - dblLatencyPeak (float): timestamp corresponding to peak
        - tplRestrictRange [2 x 1]: restrict peak to lie within tplRestrictRange(1) and tplRestrictRange(end)

    Output:
        dOnset: dict with entries:
        - dblLatencyPeakOnset: time of peak onset (first crossing half-height of peak)
        - dblValue: value at peak onset
        - dblBaseValue: baseline value (first value of tplRestrictRange, or first value in vecData if None)
        - dblLatencyPeak: time of peak
        - dblPeakValue: value of peak

    Version history:
    1.0 - June 19, 2020 Created by Jorrit Montijn, Translated to Python by Alexander Heimel
    2.0 - August 22, 2023 Updated translation by JM
    2.0.1 - August 30 2023, Bug fix for peak time definition [by JM]
    """

    # check input
    if tplRestrictRange is None:
        tplRestrictRange = [np.min(vecT), np.min(vecT)+np.ptp(vecT)]

    # remove time points outside restricted range
    indKeep = np.logical_and(vecT >= tplRestrictRange[0], vecT <= tplRestrictRange[1])
    vecCropT = vecT[indKeep]
    vecDataCropped = vecData[indKeep]

    # find peak if none supplied
    if dblLatencyPeak is None:
        intPeakIdx = np.argmax(vecDataCropped)
        dblLatencyPeak = vecCropT[intPeakIdx]
    else:
        intPeakIdx = findfirst(vecCropT >= dblLatencyPeak)
        if intPeakIdx is None:
            logging.warning(
                "getOnset:InvalidPeakTime: Supplied peak was invalid; taking max value of cropped data")
            intPeakIdx = np.argmax(vecDataCropped)
            dblLatencyPeak = vecCropT[intPeakIdx]

    dblPeakValue = vecDataCropped[intPeakIdx]

    # calculate first timepoint crossing half-height of peak
    dblBaseValue = vecDataCropped[0]
    dblThresh = (dblPeakValue - dblBaseValue)/2 + dblBaseValue
    if dblThresh > 0:
        intOnsetIdx = findfirst(vecDataCropped >= dblThresh)
    else:
        intOnsetIdx = findfirst(vecDataCropped <= dblThresh)

    if intOnsetIdx is None:
        dblLatencyPeakOnset = np.nan
        dblValue = np.nan
    else:
        dblLatencyPeakOnset = vecCropT[intOnsetIdx]
        dblValue = vecDataCropped[intOnsetIdx]

    # return
    dOnset = dict()
    dOnset['dblLatencyPeakOnset'] = dblLatencyPeakOnset
    dOnset['dblValue'] = dblValue
    dOnset['dblBaseValue'] = dblBaseValue
    dOnset['dblLatencyPeak'] = dblLatencyPeak
    dOnset['dblPeakValue'] = dblPeakValue

    return dOnset



import scipy.io
from zetapy import getZeta
import os
import numpy as np
import matplotlib.pyplot

"""Run example ZETA-test

This code loads data from an example LP cell and performs a ZETA-test,
makes a raster plot and calculates the instantaneous firing rate

Version history:
1.0 - 17 June 2020
1.1 - 23 Feb 2021 Refactoring of python code by Guido Meijer
Created by Jorrit Montijn, translated to python by Alexander Heimel
"""

# set random seed
np.random.seed(1)

# load data for example cell
strDataFile = os.path.join(os.path.dirname(__file__), 'ExampleDataZETA.mat')
dLoad = scipy.io.loadmat(strDataFile)

# some information about the neuron is stored in the dNeuron structure,
# such as whether Kilosort2 thought it was an acceptable neuron
dNeuron = dLoad['sNeuron']
if dNeuron['KilosortGood'] == 0 or dNeuron['NonStationarity'] > 0.5:
    raise(Exception('BadUnit: This unit is non-stationary, noise-like, or contaminated'))

# retrieve the spike times as an array from the field in dNeuron
arrSpikeTimes = dNeuron['SpikeTimes'][0][0]

# load stimulation information
sStim = dLoad['sStim']
arrStimulusStartTimes = sStim['StimOnTime'][0][0][0]  # unpacking Matlab array
arrStimulusStopTimes = sStim['StimOffTime'][0][0][0]  # unpacking Matlab array

# put stimulus start and stop times together into a [T x 2] matrix
arrEventTimes = np.transpose(np.array([arrStimulusStartTimes,arrStimulusStopTimes]))

# run the ZETA-test with default parameters
dblZetaP, vecLatencies = getZeta(arrSpikeTimes, arrEventTimes)

print(f'\nDefault parameters\np-value: {dblZetaP}\nlatencies: {vecLatencies}')

## run the ZETA-test with specified parameters

# median of trial-to-trial durations
dblUseMaxDur = np.median(np.diff(arrStimulusStartTimes))

# 50 random resamplings should give us a good enough idea if this cell is responsive.
# If it's close to 0.05, we should increase this number.
intResampNum = 50

# what do we want to plot?(0=nothing, 1=inst. rate only, 2=traces only, 3=raster plot as well,
# 4=adds latencies in raster plot)
intPlot = 1

# how many latencies do we want? 1=ZETA, 2=-ZETA, 3=peak, 4=first crossing of peak half-height
intLatencyPeaks = 4

# do we want to restrict the peak detection to for example the time during stimulus?
# Then put (0 1) here.
tplRestrictRange = (0, np.inf)

# then run ZETA with those parameters
dblZetaP, vecLatencies = getZeta(arrSpikeTimes, arrEventTimes,
                                 dblUseMaxDur=dblUseMaxDur,
                                 intResampNum=intResampNum,
                                 intPlot=intPlot,
                                 intLatencyPeaks=intLatencyPeaks,
                                 tplRestrictRange=tplRestrictRange)


print(f'\nSpecified parameters\np-value: {dblZetaP}\nlatencies: {vecLatencies}')

